/******************************************************************************
 * Filename: bblock.c
 *
 * Created: 2011/11/26 18:25
 * Author: Ladislav Láska
 * e-mail: laska@kam.mff.cuni.cz
 *
 ******************************************************************************/

#include <stdlib.h>
#include <assert.h>

#include "program.h"
#include "instr.h"
#include "bblock.h"
#include "debug.h"

void idis_compute_jumps(struct program *prg) {
	/* Take all sections, all instructions and look wether they can jump somewhere */

	for (int i = 0; prg->sections[i] != NULL; i++) {
		for (int j = 0; j < prg->sections[i]->instrn; j++) {
			struct instr *instr = prg->sections[i]->instrs[j];
			assert(instr);
			if (instr->def->flags & IDF_BRANCH) {
				/* Find branch address */
				int found = 0;
				for (int k = 0; instr->ops[k] != NULL; k++) {
					if (instr->ops[k]->type == OT_JUMP_REL) {
						vmaddr_t jaddr = instr->vmaddr + instr->raw_length + instr->ops[k]->val;
						struct instr *ref = idis_lookup_instr_by_addr( prg, jaddr );
						if (ref == NULL) continue;
						struct instr_llist *nitem;
						/* Add reference to the jump instruction */
						nitem = malloc(sizeof(struct instr_llist));
						nitem->instr = ref;
						nitem->next = instr->jumpsto;
						instr->jumpsto = nitem;

						/* Add reference to the jumped instruction */
						nitem = malloc(sizeof(struct instr_llist));
						nitem->instr = instr;
						nitem->next = ref->jumpedfrom;
						assert(ref);
						ref->jumpedfrom = nitem;

						found = 1;
						break;
					}
				}

				/* If no address is found, suppose computed jump.
				 * Place a dummy for now, just enough to end BB here.*/
				if (!found) {
					struct instr_llist *nitem = malloc(sizeof(struct instr_llist));
					memset(nitem, 0, sizeof(struct instr_llist));
					instr->jumpsto = nitem;
				}
			}
		}
	}
}

int _idis_bblock_addrcomp_fn( const struct bblock **a, const struct bblock **b ) {
	return ((*a)->vmaddr - (*b)->vmaddr);
}

void idis_bblock_link( struct program *prg, struct bblock *from, struct bblock *to ) {
	struct bblock_llist *tmp;

	/* Add reference to source instruction. */
	tmp = from->jumpsto;
	from->jumpsto = malloc(sizeof(struct bblock_llist));
	from->jumpsto->bblock = to;
	from->jumpsto->next = tmp;

	/* Add reference to target instruction. */
	tmp = to->jumpedfrom;
	to->jumpedfrom = malloc(sizeof(struct bblock_llist));
	to->jumpedfrom->bblock = from;
	to->jumpedfrom->next = tmp;
}

void idis_build_bblocks(struct program *prg) {
	assert(!prg->bblocks);
	int size = 16;
	prg->bblocks = malloc(sizeof(struct bblock*)*size);
	prg->bblockn = 0;

	struct bblock *bb = NULL;

	/* Build structures */
	for (int i = 0; prg->sections[i] != NULL; i++) {
		for (int j = 0; j < prg->sections[i]->instrn; j++) {
			struct instr *instr = prg->sections[i]->instrs[j];
			if (!bb) {
				/* Start basic block */
				bb = malloc(sizeof(struct bblock));
				memset(bb, 0, sizeof(struct bblock));
				bb->vmaddr = instr->vmaddr;
				bb->firsti = instr;
				bb->instrs = &prg->sections[i]->instrs[j];
				bb->instrn = 1;
				assert(bb->firsti == bb->instrs[0]);
			} else bb->instrn++;
			/* Check for jump, last instruction in section, or branch destination on next instruction*/
			if (instr->jumpsto 
				 || (prg->sections[i]->instrn-1 == j)
				 || prg->sections[i]->instrs[j+1]->jumpedfrom
				)
			{
				/* End basicblock */
				if (prg->bblockn == size - 1) {
					/* Allocate space in global list if needed */
					size *= 2;
					struct bblock **bna = realloc(prg->bblocks, sizeof(struct bblock*)*size);
					assert(bna);
					prg->bblocks = bna;
				}
				bb->lasti = instr;
				prg->bblocks[prg->bblockn++] = bb;
				bb = NULL;
			}
		}
	}

	assert(prg->bblockn);

	/* Shrink bblock array to actual size */
	struct bblock **bnf = realloc(prg->bblocks, sizeof(struct bblock*)*prg->bblockn);
	assert(bnf);
	prg->bblocks = bnf;
	size = prg->bblockn;

	/* Sort bblocks */
	qsort(prg->bblocks, prg->bblockn, sizeof(struct bblock*), (__compar_fn_t) _idis_bblock_addrcomp_fn);

	/* Build bb-graph */
	for (int i = 0; i < prg->bblockn; i++) {
		struct bblock *bb = prg->bblocks[i];

		/* If instruction jumps somewhere and is not a stub, add it. */
		if (bb->lasti->jumpsto && bb->lasti->jumpsto->instr) {
			struct bblock *jumped = idis_lookup_bblock_by_instr( prg, bb->lasti->jumpsto->instr );
			assert(jumped);
			assert(bb->lasti->jumpsto->next == NULL);
			idis_bblock_link(prg, bb, jumped);
		}

		/* If instruction is not jump, is a conditional jump, add next instruction in line. */
		if ((!(bb->lasti->def->flags & IDF_BRANCH)) || (bb->lasti->def->flags & IDF_CONDITIONAL)) {
			struct bblock *bnext = idis_lookup_bblock( prg, bb->lasti->vmaddr + bb->lasti->raw_length );
			if (bnext)
				idis_bblock_link(prg, bb, bnext);
			/* else maybe connect to some fail node? */
		}
	}
}

void idis_bblock_free( struct bblock *bb ) {
	struct bblock_llist *list;

	list = bb->jumpedfrom;
	while (list != NULL) {
		struct bblock_llist *next = list->next;
		free(list);
		list = next;
	}
	
	list = bb->jumpsto;
	while (list != NULL) {
		struct bblock_llist *next = list->next;
		free(list);
		list = next;
	}

	free(list);
}

struct bblock *idis_lookup_bblock( struct program *prg, vmaddr_t vmaddr ) {
	struct bblock dummy, *dummyptr = &dummy;
	memset(&dummy,0,sizeof(struct bblock));
	dummy.vmaddr = vmaddr;
	struct bblock **bb = bsearch
		( 
			&dummyptr, prg->bblocks, prg->bblockn, 
			sizeof(struct bblock*), (__compar_fn_t) _idis_bblock_addrcomp_fn
		);
	if (!bb) return NULL;
	assert(dummy.vmaddr == (*bb)->vmaddr);
	return (*bb);
}

struct bblock *idis_lookup_bblock_by_instr( struct program *prg, struct instr *instr ) {
	assert(instr);
	return idis_lookup_bblock( prg, instr->vmaddr );
}
