/******************************************************************************
 * Filename: disasm.c
 * Description:
 *
 * Version: 1.0
 * Created: Sep 11 2011 20:22:00
 * Last modified: Sep 11 2011 20:22:00
 *
 * Author: Ladislav Láska
 * e-mail: ladislav.laska@gmail.com
 *
 ******************************************************************************/


#define _XOPEN_SOURCE 500

#include <stdio.h>
#include <stdlib.h>
#include <gelf.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>
#include <assert.h>


#include "disasm.h"
#include "program.h"
#include "error.h"
#include "debug.h"
#include "messages.h"
#include "constants.h"
#include "instr.h"
#include "util.h"
#include "symbol.h"
#include "bblock.h"
#include "assumption.h"

#include "../config.h"

char idis_next_byte_i(unsigned char **from, struct instr *instr) {
	return (instr->raw[instr->raw_length++]) = *((*from)++);
}

long long int idis_read_bytes_i(unsigned char **from, struct instr *instr, int n, int extend) {
	assert (n <= 8);
	long long int r = 0;
	for (int i = 0; i < n; i++) {
		unsigned char b = idis_next_byte_i(from, instr);
		/* Little endian! */
		int off = 8*i;
		r |= b << off;
	}
	if ((r & (1 << (n*8-1))) && extend)
		r |= -1 << (n*8);
	return r;
}

struct instr *idis_decode_instruction( unsigned char **ptr ) {
	int fail = 0;
	/* Local macro for getting modrm field */
	#define NEED_MODRM \
		do { if (!have_modrm) { instr->modrm = idis_next_byte_i(ptr, instr); have_modrm=1; } } while (0);
	int have_modrm = 0;
	
	/* Prepare field for instruction */
	struct instr *instr = alloc_instr();
	if (instr == NULL) return NULL;

	/* Read instr->prefixes, until non-prefix is seen */
	unsigned char b;
	for (;;) {
		b = idis_next_byte_i(ptr, instr);
		assert(instr->raw_length < 16);
		/* Check for REX: 0100|wrxb */
		if ((instr->prefixes & (PF_REX | PF_0F)) == 0 && ((b & 0xf0) == 0x40)) {
			instr->prefixes |= PF_REX;
			if (b & 0x8) instr->prefixes |= PF_REX_W;
			if (b & 0x4) instr->prefixes |= PF_REX_R;
			if (b & 0x2) instr->prefixes |= PF_REX_X;
			if (b & 0x1) instr->prefixes |= PF_REX_B;
			goto cont;
		}
		int p = 0;
		while (_prefix_transl[p][1] != 0) {
			if (_prefix_transl[p][1] == b) {
				instr->prefixes |= _prefix_transl[p][0];
				goto cont;
			}
			p++;
		}

		/* Have non-prefix */
		break;

		cont:
		continue;
	}

	/* Do instruction lookup */

	/* Pick correct table for single or multibyte version */
	struct instr_def ***tbl;
	if (instr->prefixes & PF_0F) 
		tbl = _opcds0f;
	else 
		tbl = _opcds;

	/* Have correct table, look at opcode and check, if it is sufficiently defined. */
	if (tbl[b][0] == NULL) { /* Invalid opcode */
		warnf("Unknown instruction opcode 0x%x encountered. Terrible things may and WILL happen!", b);
		goto invd;
	}

	if (tbl[b][1] == NULL) {/* Single instruction, no further specification needed. */
		instr->def = tbl[b][0];
	} else {				/* Multiple instructions under same opcode, examine it */
		if (tbl[b][0]->extension != NO_EXT) {
			/* Instruction needs ModR/M (REG) opcode extension */
			NEED_MODRM;
			int ext = idis_parse_modrm(instr, FI_REG) & 0x7;
			for (int i = 0; tbl[b][i] != NULL; i++) {
				if (tbl[b][i]->extension == ext) {
					instr->def = tbl[b][i];
					break;
				}
			}
			if (instr->def == NULL) {
				warnf("Instruction 0x%x requireds ModR/M extension %i, but none was found.", b, ext);
				goto invd;
			}
		} else {
			/* Does it need secondary opcode? */
			if (tbl[b][0]->opcd2 != 0) {
				/* Instruction needs secondary opcode */
				warnf("Unsupported instruction 0x%x detected (needs secondary opcode). Terrible things WILL happen!", b);
				goto invd;
			} 

			/* Does it need operand-size resolution? (TODO: Which instructions? Only 0x98 and 0x99 it seems) */
			if (((b == 0x98) || (b == 0x99)) && (!(instr->prefixes & PF_0F))) {
				int bits = 0;
				if (instr->prefixes & PF_REX) bits = T_qp;
				else if (instr->prefixes & PF_OPSIZE) bits = T_wo;
				else bits = T_do;
				
				assert(bits != 0);
				for (int i = 0; tbl[b][i] != NULL; i++) {
					if (tbl[b][i]->ops[0]->type == bits) { 
						instr->def = tbl[b][i];
						goto decode;
					}
				}
				assert(instr->def != NULL);
			}

			warnf("Bailing out for opcode %02x for unknown reason.",b);
			goto invd;
		}
	}

	decode:;
	
	/* Decode arguments: */
	int opn;

	for (opn = 0; instr->def->ops[opn] != NULL; opn++) {
		_op_def_t *op = instr->def->ops[opn];
		int regoff = idis_get_regoff(instr, op);
		switch (op->addr) {
			case A_def:
				/* Operand is fixed for this instruction. */
				switch (op->type) {
					case T_STACK:
						idis_push_operand(instr, OT_STACK, 0);
						break;
					case T_CONST:
						idis_push_operand(instr, OT_IMMEDIATE, op->def);
						break;
					default:
						warnf("Unknown implicit operand type %i for instruction %x%x.", op->type, instr->def->x0f, instr->def->opcd);
						fail = 1;
						goto end;
						idis_push_operand(instr, OT_UNKNOWN, 0);
						break;
				}
				break;
			case A_M:
			case A_E:
				/* MOD field of ModR/M specifies addressing mode */
				{
					NEED_MODRM
					int mod = idis_parse_modrm(instr, FI_MOD);
					switch (mod) {
						case 0: /* r/m specifies GPR with address, no displacement; except for rm=101 when it becomes [rip]+disp32 */
						case 1: /* r/m specifies GPR with address, 8 bit displacement */
						case 2: /* r/m specifies GPR with address, 32bit displacement */
							{
								int rm = idis_parse_modrm(instr, FI_RM);
								int bits = idis_get_addressing(instr);
								int adroff = idis_get_addroff(instr);
								int disp = 0;
								struct op* o;
								if ((mod == 0) && (rm == 0x5)) {
									long long int disp = idis_read_bytes_i(ptr, instr, 4, 1);
									if (bits == 64) {
										/* RIP + disp32 */
										idis_push_operand2(instr, OT_INDIRECT_DISP, R_RIP, disp);
									} else {
										/* disp32 */
										idis_push_operand(instr, OT_DISPLACEMENT, disp);
									}
									break;
								} else if ((rm & 0x7) == 0x4) {
									/* SIB */
									instr->sib = idis_next_byte_i(ptr, instr);
									o = idis_push_operand(instr, OT_SIB, 0);

									/* Do scale */
									switch (idis_parse_sib(instr, FI_SCALE)) {
										case 0: o->scale = 1; break;
										case 1: o->scale = 2; break;
										case 2: o->scale = 4; break;
										case 3: o->scale = 8; break;
										default: assert(0);
									}

									/* Do index */
									o->index = idis_parse_sib(instr, FI_INDEX);
									if (o->index == 0x4) o->index = R_ZERO; /* RSP is not used as index. */
									else o->index += adroff;

									/* Do base */
									o->base  = idis_parse_sib(instr, FI_BASE );
									if ((mod == 0) && ((o->base & 0x7) == 0x5)) {	
										/* RBP is never base, instead 32bit displacement is used */
										o->base = R_ZERO;
										disp = 4;
									} else {
										o->base += adroff;
										if (mod == 1) disp = 1;
										else if (mod == 2) disp = 4;
									}
								} else {
									/* Normal indirect addressing */
									if (mod == 1) disp = 1;
									else if (mod == 2) disp = 4;
									o = idis_push_operand(instr, (disp > 0) ? OT_INDIRECT_DISP : OT_INDIRECT, adroff + rm);
								}
								/* Do displacement */
								if (disp > 0) 
									o->disp = idis_read_bytes_i(ptr, instr, disp, 1);
							}
							break;
						case 3:	/* r/m specifies GPR */
							/* assert(op->addr != A_M); //Some instruction are not allowed to do this stuff. */
							idis_push_operand(instr, OT_REG, regoff + idis_parse_modrm(instr, FI_RM));
							break;
					}
				}
				break;
			case A_G:
				/* reg field of ModR/M specifies GPR. */
				NEED_MODRM
				idis_push_operand(instr, OT_REG, regoff + idis_parse_modrm(instr, FI_REG) );
				break;
			case A_O:
			case A_I: {
				/* Immediate value, type field specifies the size */
					int imm_size = idis_get_opsize(instr, op);
					long long int imm = idis_read_bytes_i(ptr, instr, imm_size, idis_get_opextend(instr, op));
					idis_push_operand(instr, OT_IMMEDIATE, imm);
				}
				break;
			case A_Z:
				/* No ModR/M byte, 3 least significant bits of opcode defines GPR */
				idis_push_operand(instr, OT_REG, regoff + (instr->def->opcd & 0x7) );
				break;
				break;
			case A_J: {
					int imm_size = idis_get_opsize(instr, op);
					long long int imm = idis_read_bytes_i(ptr, instr, imm_size, idis_get_opextend(instr, op));
					idis_push_operand(instr, OT_JUMP_REL, imm);
				}
				break;
			case A_REG:
				idis_push_operand(instr, OT_REG, regoff + op->def);
				break;
			default:
				debugf("Addressing: %i", op->addr);
				warnf("Unknown addressing mode for instruction %x, terrible things ARE happening!", b);
				goto invd;
		}
	}
	#undef NEED_MODRM

	end:
	if (fail == 0) return instr;

	invd:
	instr->def = &_invd;
	(*ptr) -= instr->raw_length-1;
	instr->raw_length = 1;
	return instr;

}

void idis_print_instruction( struct program *prg, struct instr *instr, unsigned int flags) {
	assert(instr);
	assert(instr->ops);
	assert(instr->def);
	
	/* Print address */
	printf("%16llx:      ", instr->vmaddr);

	/* Print decoded instruction bytes, first line */
	for (int i = 0; i < _config_instr_bytes_per_line; i++) {
		if (i < instr->raw_length)
			printf("%02x ", instr->raw[i] & 0xff);
		else printf("   ");
	}

	printf("  %s", strinstr(instr));

	/* Resolve symbols for jump instructions*/
	if (instr->def->flags & IDF_BRANCH) {
		struct op *op;
		for (int i = 0; (op = instr->ops[i]) != NULL; i++) {
			if (op->type == OT_JUMP_REL) {
				vmaddr_t jaddr = instr->vmaddr + instr->raw_length + (signed long long) op->val;
				struct symbol *symb = idis_lookup_symbol_nearest( prg, jaddr);
				if (symb != NULL) {
					vmaddr_t joff = jaddr - symb->addr;
					if (joff)	printf(" <%s%s>", symb->name, strrelf(joff, IFMT_ATT));
					else		printf(" <%s>", symb->name);
				}
				break;
			}
		}
	}

	/* Print decoded instruction bytes, second line if needed */
	if (_config_instr_bytes_per_line < instr->raw_length)
	for (int i = _config_instr_bytes_per_line; i < 16; i++) {
		if ((i % _config_instr_bytes_per_line) == 0) {
			if (i >= instr->raw_length) break;
			else printf("\n%16llx:      ", instr->vmaddr+i);
		}
		if (i < instr->raw_length)
			printf("%02x ", instr->raw[i] & 0xff);
		else printf("   ");
	}

	int comments = 0;
	for (int i = 0; i < instr->userdatan; i++) {
		struct userdata *d = instr->userdata[i];
		if (d->type == UDT_COMMENT) {
			printf("\t\t\t# %s\n", (char *) d->data);
			comments++;
		}
	}

	if (comments == 0) printf("\n");
}

void idis_disassemble_section(struct section *sec) {
	assert( sec->instrs == NULL);
	sec->instrs = malloc(sizeof(struct instr*)*sec->data_size);
	size_t caddr = sec->addr;
	struct instr *instr;
	unsigned char *data = sec->data;
	unsigned char *data_end = sec->data + sec->data_size;
	int in = 0;
	while (data < data_end) {
		instr = idis_decode_instruction( &data );
		if (instr == NULL) {
			warnf("Instruction decoding failed at address %x, recovering...", (unsigned int) caddr);
		}
		assert(instr != NULL);
		sec->instrs[in++] = instr;
		instr->vmaddr = caddr;
		caddr += instr->raw_length;
	}
	sec->instrn = in;
	sec->instrs[in++] = NULL;
	void *nptr = realloc(sec->instrs,sizeof(struct instr*)*in);
	if (nptr != NULL) sec->instrs = nptr;
	else warn("Reallocation failed when attempting to shrink.");
	return;
}

void idis_dump_section(struct section *sec) {
	assert( sec->instrs != NULL);
	printf("\nDisassembly of section %s:\n", sec->name);
	struct bblock *bblock = NULL;
	for (int i = 0; i < sec->instrn; i++) {
		struct symbol *sym = idis_lookup_symbol(sec->prg, sec->instrs[i]->vmaddr);
		if (sym != NULL)
			printf("\n%016x <%s>:\n", (unsigned int) sym->addr, sym->name);
	
		if ((bblock == NULL) && (_config_show_bblocks)) {
			bblock = idis_lookup_bblock(sec->prg, sec->instrs[i]->vmaddr);
			if (bblock)
				printf("\n-- BBLOCK [0x%llx-%llx] -->\n", (unsigned long long) bblock->vmaddr, (unsigned long long) bblock->lasti->vmaddr);
		}

		idis_print_instruction(sec->prg, sec->instrs[i], IFMT_ATT);

		if ((bblock != NULL) && (bblock->lasti->vmaddr == sec->instrs[i]->vmaddr) && _config_show_bblocks) {
			printf("<-- BBLOCK ( ");
			/* Print why */
			int comma = 0;
			if (bblock->lasti->def->flags & IDF_BRANCH)
				comma++, printf("branch");
			struct instr_llist *li = bblock->lasti->jumpedfrom;
			while (li != NULL) {
				if (comma++) printf(", ");
				printf("jumped from 0x%llx", li->instr->vmaddr);
				li = li->next;
			}

			if (comma == 0) printf("no reason");

			printf(" )\n");
			bblock = NULL;
		}
	}
}

struct program *idis_disassemble_file(char *file) {
	/* Init elf and check for supported version */
	if (elf_version(EV_CURRENT) == EV_NONE) 
		_idis_error_message = MSG_ELF_INIT;

	/* Open file */
	int fd = open(file, O_RDONLY, 0);
	if (fd < 0) {
		_idis_error_message = strerror(errno);
		return NULL;
	}

	/* Load elf */
	Elf *e;
	if ((e = elf_begin(fd, ELF_C_READ, NULL)) == NULL) {
		_idis_error_message = elf_errmsg(-1);
		close(fd);
		return NULL;
	}

	/* Check filetype */
	if (elf_kind(e) != ELF_K_ELF) {
		_idis_error_message = MSG_ELF_NOT_ELF;
		elf_end(e);
		close(fd);
		return NULL;
	}
	
	/* Initialize structure */
	struct program *prg = idis_alloc_program();
	prg->elf = e;
	prg->file = file;

	/* Read executable header */
	prg->elfclass = gelf_getclass(e);
	if (prg->elfclass == ELFCLASSNONE) warn("Uknown ELF class.");
	debugf("Found %ibit elf object.", (prg->elfclass == ELFCLASS32) ? 32 : 64);

	/* Scan program header */
	size_t prgn;
	if (elf_getphdrnum(e, &prgn) != 0) {
		warn("Could not load program header. Will try to recover.");
	}
			
	GElf_Phdr phdr;
	for (int i = 0; i < prgn; i++) {
		if (gelf_getphdr(e, i, &phdr) != &phdr) 
			warnf("Could not read program header entry #%i. Will try to recover.", i);
	}

	/* Scan sections */
	Elf_Scn *scn = NULL;
	Elf_Data *scn_data;
	size_t shstrndx;
	char *scn_name;
	GElf_Shdr shdr;

	if (elf_getshdrstrndx(e, &shstrndx) != 0) {
		_idis_error_message = elf_errmsg(-1);
		elf_end(e);
		close(fd);
		return NULL;
	}

	/* Count sections */
	int scnn = 0;
	while ((scn = elf_nextscn(e, scn)) != NULL)
		scnn++;

	/* Allocate space for sections */
	prg->sections = malloc(sizeof(struct section*)*(scnn+1));
	prg->sections[scnn] = NULL;
	prg->secn = scnn;

	scn = NULL;
	int scni = 0;
	while ((scn = elf_nextscn(e, scn)) != NULL) {
		/* Read headers from section */
		if (gelf_getshdr(scn, &shdr) != &shdr) {
			_idis_error_message = MSG_ELF_SECTION_HEADER;
			/* TODO: More cleanup? */
			elf_end(e);
			close(fd);
			return NULL;
		}

		if ((scn_name = elf_strptr(e, shstrndx, shdr.sh_name)) == NULL) {
			warn("Cannot read section name!");
			scn_name = "[unknown]";
		}
		scn_name = strdup(scn_name);
		
		/* Allocate section descriptor */
		struct section *sec = malloc(sizeof(struct section));
		memset(sec, 0, sizeof(struct section));
		prg->sections[scni++] = sec;
		sec->prg = prg;
		sec->name = scn_name;
		sec->addr = shdr.sh_addr;
		sec->type = shdr.sh_type;
		if (shdr.sh_flags & SHF_EXECINSTR) sec->flags |= SF_EXE;

		/* Read contents of section */
		scn_data = NULL;
		while ((scn_data = elf_getdata(scn, scn_data)) != NULL) {
			assert(sec->data == NULL);
			if (scn_data->d_size == 0) continue;

			if (sec->addr > 0) {
				debugf("Reading section data: %s", sec->name);
				/* Copy data from Elf_Data to our structure. */
				sec->data_size = scn_data->d_size;
				sec->data = malloc(sec->data_size);
				if (scn_data->d_buf != NULL) {
					memcpy(sec->data, scn_data->d_buf, scn_data->d_size);
				} else {
					/* Program data at given mem offset, but no predefined values */
					idis_memset32(sec->data, 0xbaddcafe, scn_data->d_size);
					assert(shdr.sh_type == SHT_NOBITS);
				}
			}

			if ((sec->type == SHT_SYMTAB)
			   || (sec->type == SHT_RELA)) {
				/* Read symbol table */
				int symbs = shdr.sh_size / shdr.sh_entsize;
				debugf("Found %i symbols.", symbs);

				/* Allocate space pointers */
				int soff = prg->symbn;
				struct symbol **s = realloc(prg->symbs, (soff+symbs)*sizeof(struct symbol*));
				if (s == NULL) {
					warn("Could not malloc() space for symbols.");
					continue;
				}
				prg->symbs = s;
				prg->symbn = soff+symbs;

				for (int i = 0; i < symbs; i++) {
					if (sec->type == SHT_SYMTAB) {
						GElf_Sym sym;
						gelf_getsym(scn_data, i, &sym);
						struct symbol *ns = malloc(sizeof(struct symbol));
						ns->name = strdup(elf_strptr(e, shdr.sh_link, sym.st_name));
						ns->addr = sym.st_value;
						s[soff+i] = ns;
						//debugf("Found symbol (%i): %s at 0x%x.", i, ns->name, ns->addr);
					} else if (sec->type == SHT_RELA) {

						GElf_Rela rel;
						gelf_getrela(scn_data, i, &rel);
						
						/* shdr.sh_link is index of dynsym section, which contains symbols. */
						Elf_Scn *dynsym;
						dynsym = elf_getscn( e, shdr.sh_link );
						assert(dynsym);

						Elf_Data *dynsym_data = NULL;
						GElf_Shdr dynsym_shdr;
						gelf_getshdr( dynsym, &dynsym_shdr );

						dynsym_data = elf_getdata( dynsym, dynsym_data );
						assert(dynsym_data);

						/* Now fetch symbol */
						GElf_Sym sym;
						gelf_getsym(dynsym_data, GELF_R_SYM(rel.r_info), &sym);

						struct symbol *ns = malloc(sizeof(struct symbol));
						char *tn = elf_strptr(e, dynsym_shdr.sh_link, sym.st_name);
						assert(tn);
						ns->name = malloc(strlen(tn)+4+1);
						strcpy(ns->name, tn);
						ns->addr = rel.r_offset;

						s[soff+i] = ns;
					} else assert(0);
				}
			}
		}
	}

	/* Finalize symbol table */
	idis_finalize_symbols(prg);

	/* Disassemble data sections */
	for (int i = 0; i < scnn; i++) {
		struct section *sec = prg->sections[i];
		if ((sec->flags & SF_EXE) == 0) continue;
		assert(sec->data != NULL);

		/* Disassemble data */
		idis_disassemble_section(sec);
	}

	/* Compute jumps and build basic blocks */
	idis_compute_jumps(prg);
	idis_build_bblocks(prg);

	return prg;
}
	
void idis_dump_program(struct program *prg) {
	for (int i = 0; prg->sections[i] != NULL; i++) {
		if ((prg->sections[i]->flags & SF_EXE) == 0) continue;
		idis_dump_section(prg->sections[i]);
	}
}

void idis_close_program(struct program *prg) {
	if (prg->sections != NULL) {
		for (int i = 0; prg->sections[i] != NULL; i++) {
			struct section *sec = prg->sections[i];
			free(sec->name);
			if (sec->data != NULL) free(sec->data);
		}

	}
	free(prg);
}
