/******************************************************************************
 * Filename: expr.c
 *
 * Created: 2012/02/13 00:32
 * Author: Ladislav Láska
 * e-mail: laska@kam.mff.cuni.cz
 *
 ******************************************************************************/

#include <assert.h>
#include <stdio.h>
#include "expr.h"
#include "instr.h"
#include "constants.h"
#include "assumption.h"
#include "debug.h"
#include "util.h"

char *_idis_expr_opstr[] = {
	"?", "%", "$", "*", "+", "-", "*", "/", ":="
};

struct expr *idis_expr_new(int opn) {
	struct expr *ex = malloc(sizeof(struct expr));
	memset(ex, 0, sizeof(struct expr));
	ex->ops = malloc(sizeof(struct expr*)*opn);
	memset(ex->ops, 0, sizeof(struct expr*)*opn);
	ex->opn = opn;
	return ex;
}

void idis_expr_free( struct expr *ex ) {
	assert(ex->type >= __EX_OPERATOR || ex->opn == 0);
	for (int i = 0; i < ex->opn; i++)
		idis_expr_free(ex->ops[i]);
	if (ex->opn > 0) free(ex->ops);
	free(ex);
}

struct expr *idis_expr_new_value( value_t val ) {
	struct expr *ex = idis_expr_new(0);
	ex->type = EX_CONST;
	ex->val = val;
	return ex;
}

struct expr *idis_expr_new_reg( value_t reg ) {
	assert( reg >= 0 && reg < __R_LAST );
	struct expr *ex = idis_expr_new(0);
	ex->type = EX_REG;
	ex->val = reg;
	return ex;
}

struct expr *idis_expr_new_binary( unsigned short int type, 
	struct expr *op1, struct expr *op2 ) 
{
	struct expr *ex = idis_expr_new(2);
	ex->type = type;
	ex->ops[0] = op1;
	ex->ops[1] = op2;
	return ex;
}

struct expr *idis_expr_new_unary( unsigned short int type, struct expr *op ) {
	struct expr *ex = idis_expr_new(1);
	ex->type = type;
	ex->ops[0] = op;
	return ex;
}

struct expr *idis_expr_copy( struct expr *expr ) {
	struct expr *new = idis_expr_new(expr->opn);
	new->type = expr->type;
	for (int i = 0; i < expr->opn; i++)
		new->ops[i] = idis_expr_copy( expr->ops[i] );
	if (expr->opn == 0) new->val = expr->val;
	return new;
}

struct expr *idis_operand_enumerate_partial( struct op *op ) {
	switch (op->type) {
		case OT_REG:
			return idis_expr_new_reg( op->val );
		case OT_IMMEDIATE:
			return idis_expr_new_value( op->val );
		case OT_DISPLACEMENT:
			return idis_expr_new_unary( EX_DEREF,
				idis_expr_new_value( op->val ) );
		case OT_INDIRECT:
			return idis_expr_new_unary( EX_DEREF,
				idis_expr_new_reg( op->val ) );
		case OT_INDIRECT_DISP:
			return idis_expr_new_unary( EX_DEREF,
				idis_expr_new_binary( EX_PLUS, 
					idis_expr_new_reg( op->val ),
					idis_expr_new_value( op->disp )));
		case OT_JUMP_REL:
			return idis_expr_new_binary( EX_PLUS,
				idis_expr_new_reg( R_RIP ),
				idis_expr_new_value( op->val ));
		case OT_STACK:
			return NULL;
		case OT_SIB:;
			struct expr *addr;
				addr = idis_expr_new_binary( EX_PLUS,
					idis_expr_new_reg( op->base ),
					idis_expr_new_value( op->disp ));
			if (op->index != R_ZERO) {
				addr = idis_expr_new_binary( EX_PLUS,
					addr,
					idis_expr_new_binary( EX_MUL,
						idis_expr_new_reg( op->index ),
						idis_expr_new_value( op->scale )));
			}
			return idis_expr_new_unary( EX_DEREF, addr );
		default:
			assert(0);
			return NULL;
	}
}

/* 
 * Apply assumptions to the expression. All assumptions will be applied BEFORE 
 * execution of instr.
 */
struct expr *idis_expr_apply_assumptions( struct expr *expr, struct program *prg, struct instr *instr ) {
	if (expr == NULL) return NULL;
	switch (expr->type) {
		case EX_REG:;
			struct assumption *a = idis_assumption_getfor(prg, instr, AK_REG, expr->val );
			if (a != NULL && a->value_type == AV_IMM) {
				return idis_expr_new_value( a->value.imm );
			} else return idis_expr_copy(expr);
		case EX_PLUS:
		case EX_MINUS:
		case EX_MUL:
		case EX_DIV:
		case EX_DEREF:;
			struct expr *new = idis_expr_new( expr->opn );
			new->type = expr->type;
			new->opn = expr->opn;
			for (int i = 0; i < expr->opn; i++) 
				new->ops[i] = idis_expr_apply_assumptions( expr->ops[i], prg, instr );
			return new;
			break;
		case EX_CONST:
		case EX_ASSIGN:
			return idis_expr_copy(expr);
		default:
			assert(0);
	}
}

/*
 * Enumerates expression at given instruction in program, if there is enough 
 * information.
 *
 * Returns 0 if unsuccessfull and idis_expr_error() shall be called to detect 
 * wether or not the result was just zero.
 *
 * NOTE: Use wrapper function idis_expr_enumerate() unless you know what you're 
 * doing.
 */
void _idis_expr_enumerate( struct expr *ex, struct program *prg, struct instr *instr ) {
	if (ex == NULL) return;
	int nc = 0;
	if (ex->type == EX_CONST) return;
	else {
		/* First, traverse all operands and enumerate them. */
		for (int i = 0; i < ex->opn; i++) {
			if (ex->type != EX_CONST) {
				_idis_expr_enumerate( ex->ops[i], prg, instr );
				if (ex->ops[i]->type != EX_CONST) nc++;
			}
		}

		/* Collapse all constants into one */
		int lastconst = -1;
		for (int i = 0; i < ex->opn; i++) {
			/* Find first constant, leave others to their's own destiny. */
			if (lastconst == -1) {
				if (ex->ops[i]->type == EX_CONST) lastconst = i;
				continue;
			}
			
			/* Fold all following to lastconst */
			if (ex->ops[i]->type == EX_CONST) {
				switch (ex->type) {
					case EX_PLUS:
						ex->ops[lastconst]->val += ex->ops[i]->val;
						goto do_free;
					case EX_MINUS:
						ex->ops[lastconst]->val -= ex->ops[i]->val;
						goto do_free;
					case EX_MUL:
						ex->ops[lastconst]->val *= ex->ops[i]->val;
						goto do_free;
					case EX_DIV:
						ex->ops[lastconst]->val /= ex->ops[i]->val;
						goto do_free;
					default:
						assert(0);
				}

				do_free:
					idis_expr_free( ex->ops[i] );
					ex->ops[i] = NULL;
			} else {
				lastconst = -1;
				continue;
			}
		}

		/* Shift operands to compact interval. */
		int newlen = 0;
		for (int i = 0; i < ex->opn; i++) {
			if (ex->ops[i] == NULL) { 
				for (int j = i; j < ex->opn; j++) {
					if (ex->ops[j] != NULL) {
						ex->ops[i] = ex->ops[j];
						ex->ops[j] = NULL;
						newlen++;
						goto next;
					}
				}
			} else {
				newlen++;
				continue;
			}
			next:;
		}
		ex->opn = newlen;

		/* Erase binary operands if useless. */
		if (ex->opn == 1) 
		switch (ex->type) {
			case EX_PLUS:
			case EX_MINUS:
			case EX_MUL:
			case EX_DIV:;
				struct expr *old = ex->ops[0];
				memcpy(ex, old, sizeof(struct expr));
				free(old);
			default:;
		}
	}
	return;
}

int _idis_expr_enumerate_error = 0;

/* 
 * Wrapper for function above. This creates a working copy and then releases it. 
 */
value_t idis_expr_enumerate( struct expr *expr, struct program *prg, struct instr *instr ) {
	if (expr == NULL) {
		_idis_expr_enumerate_error = 0;
		return 0;
	}
	struct expr *ex = idis_expr_apply_assumptions( expr, prg, instr );
	_idis_expr_enumerate( ex, prg, instr );
	if ( ex->type == EX_CONST ) {
		value_t val = ex->val;
		_idis_expr_enumerate_error = 0;
		idis_expr_free(ex);
		return val;
	} else {
		_idis_expr_enumerate_error = 1;
		idis_expr_free(ex);
		return -1;
	}
}

struct expr *idis_expr_simplify( struct expr *expr, struct program *prg, struct instr *instr ) {
	struct expr *ex = idis_expr_apply_assumptions( expr, prg, instr );
	_idis_expr_enumerate( ex, prg, instr );
	return ex;
}
 

/*
 * Returns 1 if there was enumeration error.
 */
int idis_expr_enumerate_error() {
	return _idis_expr_enumerate_error;
}

/* 
 * Returns newly allocated string representation of expression. 
 */
char *_idis_strexpr( struct expr *ex, int parenthesis ) {
	/* TODO: Something more efficient. */

	if (ex->type == EX_REG) {
		return strdup(strreg(ex->val, IFMT_ATT));
	} else if (ex->type == EX_CONST) {
		char _buf[32];
		snprintf(_buf, 32, "%s", strimm(ex->val));
		return strdup(_buf);
	} else if (ex->type == EX_DEREF) {
		char *op = idis_strexpr(ex->ops[0]);
		char *buf = malloc(strlen(op)+3);
		sprintf(buf, "[%s]", op);
		free(op);
		return buf;
	} else if (ex->type < __EX_BINARY) {
		char *op = idis_strexpr(ex->ops[0]);
		char *buf = malloc(strlen(op)+strlen(_idis_expr_opstr[ex->type])+1);
		sprintf(buf, "%s%s", _idis_expr_opstr[ex->type], op);
		free(op);
		return buf;
	} else {
		char **opstr = malloc(sizeof(char *)*ex->opn);
		int len = 0;
		for (int i = 0; i < ex->opn; i++) {
			opstr[i] = _idis_strexpr(ex->ops[i], 1);
			len += strlen(opstr[i]);
		}
		/* Operators + spaces + brackets + \0 */
		len += strlen(_idis_expr_opstr[ex->type])*(len-1) + 2*(len-1) + 3; 
		char *buf = malloc(len);
		int pos = 0;
		if (parenthesis)
			pos += sprintf(buf+pos, "(%s", opstr[0]);
		else
			pos += sprintf(buf+pos, "%s", opstr[0]);
		free(opstr[0]);
		for (int i = 1; i < ex->opn; i++) {
			if ((ex->type == EX_PLUS) && (opstr[i][0] == '-')) 
				pos += sprintf(buf+pos, " - %s", opstr[i]+1);
			else
				pos += sprintf(buf+pos, " %s %s", _idis_expr_opstr[ex->type], opstr[i]);
			free(opstr[i]);
		}
		if (parenthesis)
			pos += sprintf(buf+pos, ")");
		free(opstr);
		return buf;
	}
	return strdup("?UNDEFINED!");
}

char *idis_strexpr( struct expr *ex ) {
	return _idis_strexpr( ex, 0 );
}

int idis_expr_compare( struct expr *ex1, struct expr *ex2 ) {
	if (ex1 == NULL) return 1;
	if (ex2 == NULL) return -1;
	if (ex1 == ex2) return 0;

	/* Compare type */
	if (ex1->type != ex2->type) {
		return ex1->type - ex2->type;
	}

	if (ex1->opn != ex2->opn) {
		return ex1->opn - ex2->opn;
	}

	/* Compare registers and constants */
	if (ex1->type == EX_REG || ex1->type == EX_CONST) {
		return ex1->val - ex2->val;
	}

	assert(ex1->opn);

	/* Compare operands */
	for (int i = 0; i < ex1->opn; i++) {
		int comp = idis_expr_compare(ex1->ops[i], ex2->ops[i]);
		if (comp) return comp;
	}

	return 0;
}

int idis_expr_equal( struct expr *ex1, struct expr *ex2 ) {
	return (idis_expr_compare(ex1, ex2) == 0) ? !0 : 0;
}

/* 
 * Simple hashing function, inspired by djb2.
 */
unsigned int idis_expr_hash( struct expr *ex ) {
	int h = ex->type;
	if (ex->opn == 0) h += ex->val * 33;
	else for (int i = 0; i < ex->opn; i++) 
		h = h*33 + idis_expr_hash(ex->ops[i]);
	return h;
}
