/******************************************************************************
 * Filename: instr.c
 *
 * Created: 2011/10/16 21:56
 * Author: Ladislav Láska
 * e-mail: laska@kam.mff.cuni.cz
 *
 ******************************************************************************/

#include "python.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include "instr.h"
#include "constants.h"
#include "debug.h"
#include "util.h"
#include "symbol.h"
#include "expr.h"

/* This defines global instr array */
#include "instr_data.h"

int _prefix_transl[][2] = {
	{PF_NULL, 0x26},
	{PF_NULL, 0x2e},
	{PF_NULL, 0x36},
	{PF_NULL, 0x3e},
	{PF_FS, 0x64},
	{PF_ALTER, 0x64},
	{PF_GS, 0x65},
	{PF_OPSIZE, 0x66},
	{PF_ADDRSIZE, 0x67},
	{PF_LOCK, 0xf0},
	{PF_REPNZ, 0xf2},
	{PF_REPZ, 0xf3},
	{PF_0F, 0x0f},
	{0}
};

/* Invalid instruction placeholder. */
struct instr_def _invd = {
	.mnem = "<invalid>",
	.desc = "Invalid instruction.",
	.prefix = 0x0,
	.opcd = 0,
	.opcd2 = 0,
	.extension = 8,
	.flags = 0,
	.mask = 0xff,
	.ops = NULL
};

struct instr *alloc_instr() {
	struct instr *instr = malloc(sizeof(struct instr));
	memset(instr, 0, sizeof(struct instr));

	instr->raw = malloc(MAX_INSTR_LENGTH);
	memset(instr->raw, 0, MAX_INSTR_LENGTH);
	
	instr->ops = malloc(sizeof(struct op*)*(MAX_OPERANDS+1));
	memset(instr->ops, 0, sizeof(struct op*)*(MAX_OPERANDS+1));
	return instr;
}

void free_instr( struct instr *instr ) {
	if (instr->raw != NULL) free(instr->raw);
	if (instr->ops != NULL) free(instr->ops);
	for (int i = 0; i < MAX_OPERANDS+1; i++) {
		if (instr->ops[i] == NULL) break;
		free(instr->ops[i]);
	}
	/* In future, this should also deallocate all auxilary space in op */
	free(instr);
}

unsigned char idis_parse_modrm( struct instr* instr, int field ) {
	switch (field) {
		case FI_MOD:
			return (instr->modrm & 0xc0) >> 6;
		case FI_REG:;
			unsigned char r = (instr->prefixes & PF_REX_R) ? 0x8 : 0;
			return ((instr->modrm & 0x38) >> 3) | r;
		case FI_RM:;
			unsigned char b = (instr->prefixes & PF_REX_B) ? 0x8 : 0;
			return (instr->modrm & 0x7) | b;
		default: 
			warn("Unknown ModR/M field requested.");
			assert(0);
	}
}

unsigned char idis_parse_sib( struct instr* instr, int field ) {
	switch (field) {
		case FI_SCALE:
			return (instr->sib & 0xc0) >> 6;
		case FI_INDEX:;
			unsigned char i = (instr->prefixes & PF_REX_X) ? 0x8 : 0;
			return ((instr->sib & 0x38) >> 3) | i;
		case FI_BASE:;
			unsigned char b = (instr->prefixes & PF_REX_B) ? 0x8 : 0;
			return (instr->sib & 0x7) | b;
		default: 
			warn("Unknown SIB field requested.");
			assert(0);
	}
}

struct op *idis_push_operand2( struct instr *instr, int type, long long int val, long long int disp ) {
	assert(instr->opn < MAX_OPERANDS);
	/* Operands are pre-allocated */
	struct op *o = malloc(sizeof(struct op));
	memset(o, 0, sizeof(struct op));
	o->type = type;
	o->val = val;
	o->disp = disp;
	instr->ops[instr->opn++] = o;
	instr->ops[instr->opn] = NULL;
	return o;
}

struct op *idis_push_operand( struct instr *instr, int type, long long int val) {
	return idis_push_operand2( instr, type, val, 0);
}

int idis_get_regoff(struct instr *instr, struct op_def *op ) {
	int opbits = idis_get_opsize(instr, op);

	switch (opbits) {
		case BITS8:
			if (instr->prefixes & PF_REX) return _REG8_REX;
			else return _REG8;
		case BITS16:	return _REG16;
		case BITS32:	return _REG32;
		case BITS64:	return _REG64;
	}
	return 0;
}

int idis_get_addroff(struct instr *instr) {
	if (instr->prefixes & PF_ADDRSIZE) 
		return _REG32;
	else return _REG64;
}

int idis_get_addressing(struct instr *instr) {
	if (instr->prefixes & PF_ADDRSIZE) 
		return 32;
	else return 64;
}

int idis_get_opsize(struct instr *instr, struct op_def *op) {
	switch (op->type) {
		case T_b:
		case T_bs:
		case T_bss:
			return BITS8;
		case T_CONST:
		case T_qp:
		case T_q:
			return BITS64;
		case T_dqp:
			if (instr->prefixes & PF_REX_W) return BITS64;
			else return BITS32;

		case T_vq: /* By default quadword, word on operand-size prefix */
			if (instr->prefixes & PF_OPSIZE) return BITS16;
			else return BITS64;
		case T_vqp:
			/* quadword if REX_W present, same as vds otherwise */
			if (instr->prefixes & PF_REX_W) return BITS64;	
		case T_vds: /* TODO: Correct? */
			return ((instr->prefixes & PF_OPSIZE) && (!(instr->prefixes & PF_REX_W))) ? BITS16 : BITS32;
		case T_v:
		case T_vs: /* Word or double word, sign extended to stack pointer */
			return ((instr->prefixes & PF_OPSIZE) ? BITS16: BITS32);
		case T_d:
			return BITS32;
		case T_def: /* This really is not true for most cases, but will suffice for now */
		case T_STACK:
			return idis_instr_basebits(instr);
		default:
			warnf("Unknown operand size %i.", op->type);
			return 0;
	}
}

int idis_get_opextend(struct instr *instr, struct op_def *op) {
	switch (op->type) {
		case T_bs:
		case T_bss:
		case T_vs:
			return 1;
		case T_vds:
			return (instr->prefixes & PF_REX_W) ? 1 : 0;
		default:
			return 0;
	}
}

/* Returns the base operand size */
int idis_instr_basebits(struct instr *instr) {
	int bits = BITS32; /* Default bits */

	/* 8bit-only instructions ignore all flags */
	if (instr->def->flags & IDF_8BIT) return BITS8;
	
	/* Instructions that have 16 or 64bit default */
	if (instr->def->flags & IDF_16BIT) bits = BITS16;
	if (instr->def->flags & IDF_64BIT) bits = BITS64;

	/* Near jumps have 64bit default */
	if (instr->def->flags & IDF_JUMP_NEAR) bits = BITS64;

	/* REX_W prefix always enables 64bit */
	if (instr->prefixes & PF_REX_W) bits = BITS64;

	/* OPSIZE prefix switches between 32 and 16 bit, is on REX_W */
	if ((bits != BITS64) && (instr->prefixes & PF_OPSIZE)) {
		if (bits == BITS16) bits = BITS32;
		if (bits == BITS32) bits = BITS16;
	}

	return bits;
}

char *strinstr(struct instr *instr) {
	assert(instr);
	return strinstrf( instr, IFMT_ATT );
}

struct instr *_idis_tostr_instr;
struct op *_idis_tostr_lastop;
int _idis_tostr_lasti;
int _idis_tostr_pass;
int _idis_tostr_i;
int _idis_tostr_att;
int _idis_tostr_format;
char _idis_tostr_buf[128];
char _idis_tostr_argbuf[128];
int _idis_tostr_pass;

char *idis_instr_tostr(struct instr *instr, int format) {
	assert(instr);

	char *ibuf = _idis_tostr_buf;
	ibuf[0] = ' ';
	ibuf[1] = 0;
	int ipos = 0;
	
	_idis_tostr_format = format;
	_idis_tostr_instr = instr;

	int att = _idis_tostr_att = (format & IFMT_ATT) ? 1 : 0;

	/* Print prefixes */
	if (instr->prefixes & PF_LOCK) ipos += sprintf(ibuf + ipos, "%s ", strprefix(PF_LOCK, format));
	if (instr->prefixes & PF_REPZ) ipos += sprintf(ibuf + ipos, "%s ", strprefix(PF_REPZ, format));
	if (instr->prefixes & PF_REPNZ) ipos += sprintf(ibuf + ipos, "%s ", strprefix(PF_REPNZ, format));

	/* Print mnemonics */
	if (att) {
		/* Copy mnemonics */
		ipos += idis_memlower(ibuf + ipos, instr->def->mnem);

		/* Prepare stuff */
		int is_branch = instr->def->flags & IDF_BRANCH;
		int is_cond = instr->def->flags & IDF_CONDITIONAL;
		UNUSED(is_cond);

		/* Search arguments */
		int has_reg = 0;
		for (int i = 0; instr->ops[i] != NULL; i++) {
			/* Skip hidden arguments */
			if (instr->def->ops[i]->flags & OF_HIDDEN) continue;
			
			if (instr->ops[i]->type == OT_REG) has_reg = 1;
		}

		/* Append suffix, if no visible register operand is found, and at least one operand is found */
		assert(instr->ops);
		int dosuffix = 1;
		
		/* Don't suffix instructions without operand and branches */
		if ((instr->ops[0] == NULL)
			|| (instr->def->flags & IDF_CONDITIONAL)
		) dosuffix = 0;

		if (has_reg & (!is_branch)) dosuffix = 0;

		if (dosuffix) {
			char sf;
			switch (idis_instr_basebits(instr)) {
				case BITS8:  sf = 'b'; break;
				case BITS16: sf = 'w'; break;
				case BITS32: sf = 'l'; break;
				case BITS64: sf = 'q'; break;
				default: sf = '?'; break;
			}
			ibuf[ipos++] = sf;
		}

	ibuf[ipos] = 0;
	} else 
		ipos += sprintf(ibuf + ipos, "  %s ", instr->def->mnem);

	/* Initialize forcycles in arg() */
	_idis_tostr_i = -1;
	_idis_tostr_pass = 0;

	return ibuf;
}

struct expr* idis_instr_tostr_arg_expr() {
	return idis_operand_enumerate_partial(_idis_tostr_lastop);
}

int idis_instr_tostr_arg_index() {
	return _idis_tostr_lasti;
}

char *idis_instr_tostr_arg() {
	int att = _idis_tostr_att;
	struct instr *instr = _idis_tostr_instr;
	char *ibuf = _idis_tostr_argbuf;
	int ipos = 0;
	int format = _idis_tostr_format;
	int have_arg = 0;

	/* Find next argument and print it. */
	if (_idis_tostr_pass < 2)
	while (_idis_tostr_i < MAX_OPERANDS+1) {
		_idis_tostr_i++;
		if (instr->ops[_idis_tostr_i] == NULL) {
			if (++_idis_tostr_pass >= 2) break;
			_idis_tostr_i = -1;
			continue;
		}
		struct op *o = _idis_tostr_lastop = instr->ops[_idis_tostr_lasti = _idis_tostr_i];
		if (instr->def->ops[_idis_tostr_i]->flags & OF_HIDDEN) continue;

		/* Skip source destinations at appropriate _idis_tostr_passes when AT&T syntax is on. */
		if (att && ((instr->def->ops[_idis_tostr_i]->flags & OF_DEST) || (instr->def->ops[_idis_tostr_i]->flags & OF_VDEST))) {
			if (_idis_tostr_pass == 0) continue;
		} else {
			if (_idis_tostr_pass == 1) continue;
		}
		
		/* Do references on computed jumps */		
		if ((instr->def->flags & IDF_BRANCH) && att) {
			switch (o->type) {
				case OT_REG:
				case OT_DISPLACEMENT:
				case OT_INDIRECT_DISP:
				case OT_SIB:
				ibuf[ipos++] = '*';
			}
		}

		/* Print operand */
		ipos += sprintf(ibuf + ipos, "%s", idis_stropf(instr, o, format));

		have_arg = 1;
		break;
	}
	if (have_arg) {
		return ibuf;
	} else {
		return NULL;
	}
}

char _idis_stropbuf[128];

char *idis_stropf(struct instr *instr, struct op *o, int format) {
	char *ibuf = _idis_stropbuf;
	int ipos = 0;
	int att = (format & IFMT_ATT) ? 1 : 0;
	switch (o->type) {
		case OT_REG:
			ipos += sprintf(ibuf + ipos, "%s", strreg(o->val, format));
			break;
		case OT_IMMEDIATE:
			if (att) ipos += sprintf(ibuf + ipos, "%s", strimmf(o->val, format));
			else 	 ipos += sprintf(ibuf + ipos, "0x%llx", o->val);
			break;
		case OT_JUMP_REL:;
			/* Translate to absolute address */
			vmaddr_t jaddr = instr->vmaddr + instr->raw_length + (signed long long) o->val;
			
			if (att) ipos += sprintf(ibuf + ipos, "%llx", jaddr);
			else 	 ipos += sprintf(ibuf + ipos, "+0x%llx", o->val);
			break;
		case OT_DISPLACEMENT:
			if (att) ipos += sprintf(ibuf + ipos, "%s()", strdispf(o->val, format));
			else 	 ipos += sprintf(ibuf + ipos, "[0x%lx]", (long int) o->val);
			break;
		case OT_INDIRECT_DISP:
			if (att) ipos += sprintf(ibuf + ipos, "%s(%s)", strdispf(o->disp, format), strreg(o->val, format));
			else 	 ipos += sprintf(ibuf + ipos, "[%s+0x%lx]", strreg(o->val, format), (long int) o->disp);
			break;
		case OT_INDIRECT:
			if (att) ipos += sprintf(ibuf + ipos, "(%s)", strreg(o->val, format));
			else 	 ipos += sprintf(ibuf + ipos, "[%s]", strreg(o->val, format));
			break;
		case OT_SIB:
			if (att) {
				/* Print displacement */
				if (o->disp != 0)
					ipos += sprintf(ibuf + ipos, "%s", strdispf(o->disp, format));
				
				if (o->index == R_ZERO)
					ipos += sprintf(ibuf + ipos, "(%s)",
						strreg(o->base, format));
				else {
					/* Do this separately, since strreg uses static string! */
					ipos += sprintf(ibuf + ipos, "(%s,",
						strreg(o->base, format));
					ipos += sprintf(ibuf + ipos, "%s,%x)", 
						strreg(o->index, format),
						o->scale);
				}
			} else {
				ipos += sprintf(ibuf + ipos, "[");
				int plus = 0;
				if (o->base != R_ZERO)
					ipos += sprintf(ibuf + ipos, "%s", strreg(o->base, format)), plus++;

				if (o->index != R_ZERO) {
					if (plus++) ipos += sprintf(ibuf + ipos, "+");
					ipos += sprintf(ibuf + ipos, "%s*%i", strreg(o->index, format), o->scale);
				}

				if (o->disp != 0) {
					if (plus++) ipos += sprintf(ibuf + ipos, "+");
					ipos += sprintf(ibuf + ipos, "0x%lx", (long int) o->disp);
				}
				ipos += sprintf(ibuf + ipos, "]");
			}
			break;
		default:
			warnf("Don't know hot to print operand type %i.", o->type);
	}
	return _idis_stropbuf;
}

char *idis_strop(struct instr *instr, struct op *op) {
	return idis_stropf(instr, op, IFMT_ATT | IFMT_IMM_SIGNED);
}

char _idis_strinstrfbuf[128];

char *strinstrf(struct instr *instr, int format) {
	char *mnem = idis_instr_tostr(instr, format);
	char *arg;
	int pos = 0;
	pos += snprintf(_idis_strinstrfbuf+pos, 128-pos, "%-4s ", mnem);
	int po = pos;
	while ((arg = idis_instr_tostr_arg()) != NULL) 
		pos += snprintf(_idis_strinstrfbuf+pos, 128-pos, "%s, ", arg);
	if (po != pos) 
		_idis_strinstrfbuf[pos-2] = 0;
	return _idis_strinstrfbuf;
}

struct op *idis_instr_dst( struct instr *instr ) {
	for (int i = 0; i < instr->opn; i++) {
		if (instr->def->ops[i]->flags & OF_DEST)
			return instr->ops[i];
	}
	return NULL;
}

struct op *idis_instr_src( struct instr *instr ) {
	for (int i = 0; i < instr->opn; i++) {
		if ((instr->def->ops[i]->flags & OF_DEST) == 0)
			return instr->ops[i];
	}
	return NULL;
}
