/******************************************************************************
 * Filename: python.c
 *
 * Created: 2012/01/12 15:53
 * Author: Ladislav Láska
 * e-mail: laska@kam.mff.cuni.cz
 *
 ******************************************************************************/

#include <Python.h>
#include <assert.h>

#include "debug.h"
#include "util.h"
#include "python.h"
#include "serializer.h"

struct python_module_llist *_idis_python_modlist = NULL;

struct python_module_llist *idis_python_lookup_module( char *module_name ) {
	struct python_module_llist *next = _idis_python_modlist;
	while (next != NULL) {
		if (strcmp(next->name, module_name) == 0)
			break;
		next = next->next;
	}
	return next;
}

/* 
 * Simple call wrapper to python.
 *
 * Note that there is wrapper idis_python_call(module_name, fn_name, ...) that counts
 * the arguments for you.
 */
int _idis_python_call(const char *module_name, const char *fn_name, int argn, ...) {
	PyObject *mod = PyImport_AddModule( module_name );
	assert(mod);
	PyObject *fn = PyObject_GetAttrString(mod, fn_name);
	if (!fn) {
		warnf("Python call failed on module %s, function %s(%i arguments): ", module_name, fn_name, argn);
		PyErr_Print();
		return -1;
	}
	assert(PyCallable_Check(fn));
	assert(argn > 0);

	va_list lst;
	va_start(lst, argn);
	
	PyObject *args = PyTuple_New(argn);
	for (int i = 0; i < argn; i++) {
		PyTuple_SetItem(args, i, va_arg(lst, PyObject*));
	}
	
	if (PyObject_CallObject(fn, args) == NULL)
		PyErr_Print();

	va_end(lst);
	
	return 0;
}


int idis_python_call_hook(struct program *prg, const char *module_name, int hook_type) {
	char *fn_name = NULL;
	switch (hook_type) {
		case PHT_INIT: fn_name = "init_hook"; break;
		case PHT_RUN: fn_name = "run_hook";	break;
		case PHT_FOR_INSTR: fn_name = "instr_hook"; break;
		case PHT_FOR_BBLOCK: fn_name = "bblock_hook"; break;
		case PHT_DISASSEMBLE: fn_name = "disassemble_hook"; break;
		case PHT_FINALIZE: fn_name = "finalize_hook"; break;
		case PHT_SERIALIZE: fn_name = "serialize_hook"; break;
		case PHT_DESERIALIZE: fn_name = "deserialize_hook"; break;
		default: assert(0);
	}
	/* Get module */
	PyObject *mod = PyImport_AddModule( module_name );
	if (!mod) {
		warnf("Module import failed when calling hook %s: %s", fn_name, module_name);
		PyErr_Print();
		return -1;
	}
	PyObject *fn = PyObject_GetAttrString(mod, fn_name);
	if (fn == NULL) { 
		PyErr_Clear();
		return 0; /* No function is OK (0 hooks were executed) */
	}
	assert(PyCallable_Check(fn));

	int hooks_called = 0;

	debugf("   %s", fn_name);
	PyObject *args;
	switch (hook_type) {
		case PHT_INIT:
			args = PyTuple_New(0);
			if (PyObject_CallObject(fn, args) == NULL) {
				debugf("Python %s failed.", fn_name);
				PyErr_Print();
				return -1;
			}
			hooks_called++;
			break;
		case PHT_SERIALIZE: 
			/* Serialize hook is special. It nees to call some helpers first! */
			idis_seri_scalar("module");
			idis_seri_mapping_start();
			idis_seri_mapping_item("handler", "module");
			idis_seri_mapping_item("name", (char *) module_name);
			idis_seri_scalar("data");
		case PHT_DESERIALIZE: /* Crawling will be done in python.i */
			args = PyTuple_New(2);
			PyTuple_SetItem(args, 0, idis_python_export_program(prg));
			PyTuple_SetItem(args, 1, idis_deser_module(prg->seri_data, (char *) module_name));
			if (PyObject_CallObject(fn, args) == NULL) {
				debugf("Python %s failed.", fn_name);
				PyErr_Print();
				return -1;
			}
			break;

		case PHT_DISASSEMBLE:
		case PHT_FINALIZE:
		case PHT_RUN:
			args = PyTuple_New(1);
			PyTuple_SetItem(args, 0, idis_python_export_program(prg));
			if (PyObject_CallObject(fn, args) == NULL) {
				debugf("Python %s failed.", fn_name);
				PyErr_Print();
				hooks_called = -1;
				goto seri_cleanup;
			}
			hooks_called++;
			seri_cleanup:
			if (hook_type == PHT_SERIALIZE)
				idis_seri_mapping_end();
			break;

		case PHT_FOR_INSTR:
			args = PyTuple_New(3);
			PyTuple_SetItem(args, 0, idis_python_export_program(prg));
			for (int s = 0; s < prg->secn; s++) {
				PyTuple_SetItem(args, 1, idis_python_export_section(prg->sections[s]));
				for (int i = 0; i < prg->sections[s]->instrn; i++) {
					struct instr *instr = prg->sections[s]->instrs[i];
					PyTuple_SetItem(args, 2, idis_python_export_instr(instr));
					if (PyObject_CallObject(fn, args) == NULL) {
						debug("Python instr_hook failed.");
						PyErr_Print();
					}
					hooks_called++;
				}
			}
			break;
		case PHT_FOR_BBLOCK:
			args = PyTuple_New(2);
			PyTuple_SetItem(args, 0, idis_python_export_program(prg));
			for (int i = 0; i < prg->bblockn; i++) {
				PyTuple_SetItem(args, 1, idis_python_export_bblock(prg->bblocks[i]));
				if (PyObject_CallObject(fn, args) == NULL) {
					debug("Python bblock_hook failed.");
					PyErr_Print();
				}
				hooks_called++;
			}
			break;
	}
	
	return hooks_called;
}

/* Call hooks for all modules */
void idis_python_call_hooks( struct program *prg, int set ) {
	struct python_module_llist *modlist = _idis_python_modlist;
	while (modlist != NULL) {
		debugf("Running hooks on module %s:", modlist->name );
		switch (set) {
			case PHS_STD:
				idis_python_call_hook( prg, modlist->name, PHT_DISASSEMBLE );
				idis_python_call_hook( prg, modlist->name, PHT_FOR_INSTR );
				idis_python_call_hook( prg, modlist->name, PHT_FOR_BBLOCK );
				idis_python_call_hook( prg, modlist->name, PHT_FINALIZE );
				break;
			case PHS_SERIALIZE:
				idis_python_call_hook( prg, modlist->name, PHT_SERIALIZE );
				break;
			case PHS_DESERIALIZE:
				idis_python_call_hook( prg, modlist->name, PHT_DESERIALIZE );
				break;
			default:
				warnf("Unknown hook set %i.", set);
		}
		modlist = modlist->next;
		debug("Done.");
	}
}

/* Loads and initializes module. */
int idis_python_load_module( char *module_name ) {
	struct python_module_llist *module_entry = idis_python_lookup_module(module_name);
	if (module_entry) {
		warnf("Module '%s'already loaded!", module_name);
		return -1;
	}

	PyImport_ImportModule( module_name );
	PyObject *mod = PyImport_AddModule( module_name );
	if (mod == NULL) {
		warnf("Could not load requested module '%s'.", module_name);
		return -1;
	}
	debugf("Running hooks on module %s:", module_name);
	idis_python_call_hook( NULL, module_name, PHT_INIT );
	module_entry = malloc(sizeof(struct python_module_llist));
	memset(module_entry, 0, sizeof(struct python_module_llist));
	module_entry->name = strdup(module_name);
	module_entry->obj = mod;
	if (_idis_python_modlist == NULL) 
		_idis_python_modlist = module_entry;
	else {
		struct python_module_llist *m = _idis_python_modlist;
		while (m->next != NULL) m = m->next;
		m->next = module_entry;
	}
	debugf("Module '%s' loaded and initialized.", module_name);
	return 0;
}

int idis_python_serialize( struct program *prg ) {
	idis_python_call_hooks( prg, PHS_SERIALIZE );
	return 0;
}

void idis_python_append_path(char *path, int autoload) {
	/* Append path */
	PyObject *mod = PyImport_AddModule("sys");
	PyObject *ppath = idis_python_export_str( path );
	PyObject *pobj = PyObject_GetAttrString( mod, "path");
	PyList_Append(pobj, ppath);
	PyErr_Print();

	if (autoload == 0) return;

	/* Load modules */
	mod = PyImport_AddModule("idis");
	PyObject *listfn = PyObject_GetAttrString( mod, "list_modules" );
	assert(listfn);
	PyObject *args = PyTuple_New(1);
	PyTuple_SetItem(args, 0, ppath);
	PyObject *modules = PyObject_CallObject(listfn, args);
	PyErr_Print();
	if (modules == NULL) {
		debugf("Path '%s' added to list, no modules found.", path);
		return;
	} else {
		int loaded = 0;
		for (int i = 0; i < PyList_Size(modules); i++) {
			PyObject *obj = PyList_GetItem(modules, i);
			char *name = PyString_AsString(obj);
			if (strcmp(name, "_gidis") == 0
				|| strcmp(name, "_idis") == 0
				|| strcmp(name, "gidis") == 0
				|| strcmp(name, "idis") == 0)
			continue;
			idis_python_load_module(name);
			loaded++;
		}
		debugf("Path '%s' added to list, %i modules autloaded.", path, loaded);
	}
}

void idis_python_init() {
	Py_Initialize();
	PyRun_SimpleString("import sys");
	PyRun_SimpleString("import os");
	PyRun_SimpleString("sys.path.insert(0, os.path.expanduser('~/.idis'))");
	PyRun_SimpleString("sys.path.insert(0, os.path.expanduser('~/.idis/plugins'))");
	PyRun_SimpleString("sys.path.insert(0, 'python')");
	PyRun_SimpleString("sys.path.insert(0, '.')");
	PyRun_SimpleString("sys.path.insert(0, '/usr/share/idis/python')");
	PyRun_SimpleString("sys.path.insert(0, '/usr/local/share/idis/python')");
	PyRun_SimpleString("import idis");
	PyRun_SimpleString("import gidis");
	PyErr_Print();
}

void idis_python_fini() {
	/* Free the whole modlist structure */
	struct python_module_llist *modlist, *next;
	modlist = _idis_python_modlist;
	while (modlist != NULL) {
		next = modlist->next;
		free(modlist->name);
		free(modlist);
		modlist = next;
	}

	Py_Finalize();
}
