/******************************************************************************
 * Filename: serializer.c
 *
 * Created: 2012/04/30 23:13
 * Author: Ladislav Láska
 * e-mail: laska@kam.mff.cuni.cz
 *
 ******************************************************************************/

#include "../core.h"
#include <stdio.h>
#include <yaml.h>
#include <string.h>

#include "serializer.h"

/* Grammar:
 *  
 * stream ::= STREAM-START document* STREAM-END
 * document ::= DOCUMENT-START node DOCUMENT-END
 * node ::= ALIAS | SCALAR | sequence | mapping
 * sequence ::= SEQUENCE-START node* SEQUENCE-END
 * mapping ::= MAPPING-START (node node)* MAPPING-END
 */

yaml_emitter_t _idis_seri_emitter;
yaml_event_t _idis_seri_event;
FILE *_idis_seri_fd;

void idis_seri_emit() {
	int ret = yaml_emitter_emit(&_idis_seri_emitter, &_idis_seri_event);
	if (ret == 0)
		warnf("YAML emit() returned error: %s", _idis_seri_emitter.problem);
}

void idis_seri_scalar(char *val) {
	debug("Scalar.");
	yaml_scalar_event_initialize(&_idis_seri_event, 
		NULL, NULL, (unsigned char*) val, strlen(val), 1, 0, YAML_ANY_SCALAR_STYLE);
	idis_seri_emit();
}

void idis_seri_mapping_start() {
	debug("Mapping start.");
	yaml_mapping_start_event_initialize(&_idis_seri_event,
		NULL, NULL, 0, YAML_ANY_MAPPING_STYLE);
	idis_seri_emit();
}

void idis_seri_mapping_end() {
	debug("Mapping end.");
	yaml_mapping_end_event_initialize(&_idis_seri_event);
	idis_seri_emit();
}

void idis_seri_mapping_item(char *key, char *value) {
	debug("Mapping item.");
	idis_seri_scalar(key);
	idis_seri_scalar(value);
}

int idis_seri_init(struct program *prg, char *filename) {
	_idis_seri_fd = fopen(filename, "w");
	if (!_idis_seri_fd) {
		_idis_error_message = strerror(errno);
		return -1;
	}
	yaml_emitter_initialize(&_idis_seri_emitter);
	yaml_emitter_set_output_file(&_idis_seri_emitter, _idis_seri_fd);
	
	/* Start document */
	yaml_stream_start_event_initialize(&_idis_seri_event, YAML_UTF8_ENCODING);
	idis_seri_emit();
	yaml_document_start_event_initialize(&_idis_seri_event, NULL,  NULL, NULL, 1);
	idis_seri_emit();

	idis_seri_mapping_start();
	idis_seri_mapping_item("file", prg->file);
	return 0;
}

int idis_seri_fini() {
	/* End document */
	idis_seri_mapping_end();
	yaml_document_end_event_initialize(&_idis_seri_event, 1);
	idis_seri_emit();
	yaml_stream_end_event_initialize(&_idis_seri_event);
	idis_seri_emit();
	yaml_emitter_delete(&_idis_seri_emitter);

	fclose(_idis_seri_fd);
	return 0;
}


/* Deserialization part */

static yaml_parser_t parser;
static yaml_event_t event;

struct seri_data *idis_deser_new() {
	struct seri_data *obj = malloc(sizeof(struct seri_data));
	memset(obj, 0, sizeof(struct seri_data));
	return obj;
}

void idis_deser_free( struct seri_data *d ) {
	while (d != NULL) {
		switch (d->type) {
			case Y_NULL: break;
			case Y_SCALAR:
				free(d->data.val);
				break;
			case Y_SEQ:
			case Y_MAP:
				idis_deser_free(d->data.obj);
				break;
		}
		struct seri_data *t = d;
		d = d->next;
		free(t);
	}
}

/* Reads next object recursively and returns it. */
struct seri_data *idis_deser_parse() {
	/* This is the node we created, event if it may be empty. */
	struct seri_data *data = idis_deser_new();
	struct seri_data *cd = data;
	struct seri_data *last = NULL;
	for (;;) {
		yaml_parser_parse(&parser, &event);
		switch (event.type) {
			case YAML_NO_EVENT: break;
			/* Ignore those, like they never happened. */
			case YAML_STREAM_START_EVENT:    break;
			case YAML_DOCUMENT_START_EVENT:  break;
			case YAML_ALIAS_EVENT:			 break;

			case YAML_SEQUENCE_START_EVENT: 
				cd->type = Y_SEQ;
				yaml_event_delete(&event);
				cd->data.obj = idis_deser_parse();
				goto advance;

			case YAML_MAPPING_START_EVENT:
				cd->type = Y_MAP;
				yaml_event_delete(&event);
				cd->data.obj = idis_deser_parse();
				goto advance;
			
			case YAML_DOCUMENT_END_EVENT:
			case YAML_STREAM_END_EVENT:
			case YAML_SEQUENCE_END_EVENT:
			case YAML_MAPPING_END_EVENT:
				yaml_event_delete(&event);
				goto end;
			
			case YAML_SCALAR_EVENT: 
				cd->type = Y_SCALAR;
				cd->data.val = strdup((char *) event.data.scalar.value);
				yaml_event_delete(&event);
				goto advance;
		}

		continue;
		advance:;
		/* Advance pointer. */
		struct seri_data *nd = idis_deser_new();
		cd->next = nd;
		last = cd;
		cd = nd;
	}
	end:
	if (last == NULL) return NULL;
	free(last->next);
	last->next = 0;
	return data;
}

void idis_deser_print( struct seri_data *d ) {
	while (d != NULL) {
		switch (d->type) {
			case Y_NULL:
				printf("NULL");
				break;
			case Y_SCALAR:
				printf("Scalar: %s\n", d->data.val);
				break;
			case Y_SEQ:
				printf("Seq...\n");
				idis_deser_print(d->data.obj);
				break;
			case Y_MAP:
				printf("Map...\n");
				idis_deser_print(d->data.obj);
				break;
		}
		assert(d);
		d = d->next;
	}
}

PyObject *idis_deser_list( struct seri_data *d ) {
	PyObject *list = PyList_New(0);
	while (d != NULL) {
		switch (d->type) {
			case Y_SCALAR:
				PyList_Append(list, PyString_FromString(d->data.val));
				break;
			case Y_SEQ:
				PyList_Append(list, idis_deser_list(d->data.obj));
				idis_deser_print(d->data.obj);
				break;
			case Y_MAP:
				PyList_Append(list, idis_deser_hash(d->data.obj));
				break;
		}
		assert(d);
		d = d->next;
	}
	return list;
}

PyObject *idis_deser_hash( struct seri_data *d ) {
	PyObject *hash = PyDict_New();
	char *key = NULL;
	while (d != NULL) {
		switch (d->type) {
			case Y_SCALAR:
				if (key == NULL) key = d->data.val;
				else {
					PyDict_SetItem(hash, PyString_FromString(key), PyString_FromString(d->data.val));
					key = NULL;
				}
				break;
			case Y_SEQ:
				assert(key != NULL);
				PyDict_SetItem(hash, PyString_FromString(key), idis_deser_list(d->data.obj));
				idis_deser_print(d->data.obj);
				key = NULL;
				break;
			case Y_MAP:
				assert(key != NULL);
				PyDict_SetItem(hash, PyString_FromString(key), idis_deser_hash(d->data.obj));
				key = NULL;
				break;
		}
		assert(d);
		d = d->next;
	}
	return hash;
}

struct seri_data *idis_deser_key( struct seri_data *d, char *key ) {
	/* Search for key */
	while (d != NULL) {
		if (((d->type == Y_SCALAR) && d->next) && (strcmp(d->data.val, key) == 0))
			return d->next;
		if (d->next == NULL) return NULL;
		/* Advance two pointers */
		d = d->next->next;
	}
	return NULL;
}

PyObject *idis_deser_module( struct seri_data *d, char *modname ) {
	while ((d = idis_deser_key(d, "module")) != NULL) {
		if (d->type != Y_MAP) continue;
		struct seri_data *sd = d->data.obj;		
		struct seri_data *name = idis_deser_key(sd, "name");
		if (name->type != Y_SCALAR) continue;
		if (name == NULL) continue;
		if (strcmp(name->data.val, modname) == 0)
			return idis_deser_hash(sd);
	}
	return NULL;
}

struct program *idis_deser( char *str ) {
	FILE *f = fopen(str, "r");
	if (!f) return NULL;

	yaml_parser_initialize(&parser);
	yaml_parser_set_input_file(&parser, f);		

	/* Load object into memory. */
	struct seri_data *d = idis_deser_parse();
	struct seri_data *file = idis_deser_key( d->data.obj, "file" );
	if (file->type != Y_SCALAR) goto error;

	struct program *prg = idis_disassemble_file( file->data.val );
	if (prg == NULL) goto error;
	
	/* First is always mapping. */
	prg->seri_data = d->data.obj;
	free(d);

	/* Good, we have program. Run hooks for it. */
	idis_python_call_hooks(prg, 0);
	idis_python_call_hooks(prg, PHS_DESERIALIZE);

	return prg;

	error:
	idis_deser_free(d);
	return NULL;
}
