/******************************************************************************
 * Filename: command.c
 *
 * Created: 2012/03/18 21:55
 * Author: Ladislav Láska
 * e-mail: laska@kam.mff.cuni.cz
 *
 ******************************************************************************/

#include "../core.h"

#include <gtk/gtk.h>
#include "view.h"
#include "dumper.h"
#include "alias.h"
#include "command.h"

struct cmd_item *_gidis_cmd_items;

#define PREFIX(str, pf) ((strncmp(str, pf, strlen(pf)) == 0) && ((str[strlen(pf)] == ' ') || (str[strlen(pf)] == 0)) )

char *gidis_command_open(struct program *prg, char *str) {
	if (prg) return M_ERR "Some program already open. Please, close it first.";
	struct program *nprg = idis_disassemble_file(str);
	if (nprg == NULL) {
		/* May be it is serialized file? */
		nprg = idis_deser( str );
		if (nprg == NULL) 
			return M_ERR "Could not disassemble file: not an idis or object file.";
	} else {
		idis_python_call_hooks(nprg, 0);
	}
	gidis_view_prg_load(nprg);
	return "Opened.";
}

char *gidis_command_close(struct program *prg) {
	if (!prg) return M_ERR "Nothing is open.";
	gidis_view_prg_free();
	idis_close_program(prg);
	return "";
}

char *gidis_command_write(struct program *prg, char *str) {
	idis_seri_init(prg, str);
	idis_python_serialize(prg);
	idis_seri_fini();
	return "Written.";
}

char _gidis_command_msg_buf[128];

char *gidis_command_alias(struct program *prg, char *str) {
	size_t alen = strcspn(str, "=");
	char *aname = str;
	char *aval = str+alen+1;
	if (str[alen] == 0)
		return M_ERR "Wrong expression: '=' character not found.";
	str[alen] = 0;

	debugf("Alias: '%s' -> '%s'", aname, aval);
	int ret = gidis_alias_new(aname, aval);
	if (ret == 0) {
		gidis_view_rebuild_linebuf();
		return "Alias inserted.";
	} else return M_ERR "Error. Probably bad syntax somewhere.";
}

char *gidis_command_search(struct program *prg, char *str) {
	GError *err = NULL;
	GRegex *expr = g_regex_new(str, 0, 0, &err);
	if (err) {
		assert(err->message);
		snprintf(_gidis_command_msg_buf, 128, "%s", err->message);
		g_error_free(err);
		return _gidis_command_msg_buf;
	}
	gidis_view_search_new();
	int count = 0;
	for (int i = 0; i < _gidis_view_instr_info_len; i++) {
		char **dump = gidis_dump_one(i);
		for (int j = 0; dump[j] != NULL; j++) {
			GMatchInfo *mi;
			g_regex_match( expr, dump[j], 0, &mi);
			while (g_match_info_matches(mi)) {
				int a,b;
				g_match_info_fetch_pos(mi, 0, &a,&b);
				count ++;
				g_match_info_next(mi, &err);
				if (err) {
					warnf("Error processing regular expression: %s", err->message);
					g_error_free(err);
					err = NULL;
					break;
				}
				gidis_view_search_mark_hi(i, GST_FULLTEXT, j, a, b);
			}
			g_match_info_free(mi);
		}
	}
	g_regex_unref(expr);

	if (err != NULL)
		g_error_free(err);
	if (count) {
		snprintf(_gidis_command_msg_buf, 128, "%i matches found.", count);
		gidis_view_search_first();
		return _gidis_command_msg_buf;
	} else
		return M_ERR "No match found.";
}

char _gidis_command_dump_buf[128];

char *gidis_command_dump(struct program *prg, char *str) {
	FILE *f = fopen(str, "w");
	if (!f) return M_ERR "dump: Could not open file.";

	int lines = gidis_dump(f, 0, -1);
	fclose(f);
	if (lines < 0)
		return M_ERR "dump: An error occurred. Dump may be incomplete or corrupt.";
	
	snprintf(_gidis_command_dump_buf, 128, "Successfully written %i lines to %s.", lines, str);
	return _gidis_command_dump_buf;
}

char _gidis_command_goaddr_buf[128];

char *gidis_command_goaddr(struct program *prg, char *str) {
	char *end;
	long long int addr = strtoll(str, &end, 16);
	debugf("Searching for address %llx\n", addr);
	if (end[0] != 0) {
		return M_ERR "Unknown command or invalid address.";
	}

	/* Find mask */
	vmaddr_t mask = 0xf;
	if (strncmp(str, "0x", 2) == 0) {
		mask = ~0;
	} else {
		for (int i = 0; i < sizeof(vmaddr_t)*2; i++) {
			if ((addr & ~mask) == 0) break;
			else mask = (mask << 4) | 0xf;
		}
	}
	debugf("Using bitmask 0x%llx", mask&addr);

	gidis_view_search_new();

	int found = 0;
	for (int i = 0; i < _gidis_view_instr_info_len; i++) {
		struct instr *instr = _gidis_view_instr_info[i].instr;
		if (instr == NULL) continue;
		if ((instr->vmaddr & mask) == addr) {
			debugf("Found address 0x%llx\n", instr->vmaddr);
			gidis_view_search_mark(i, GST_ADDR);
			found++;
		}
	}

	if (found > 0 )	{
		gidis_view_search_first();
		return NULL;
	} else return M_ERR "Unknown command or address not found.";
}

char *gidis_command_execute(struct program *prg, char *command ) {
	char *cmd = command+1;
	if (command[0] == ':') {
		int cmdlen = strcspn(cmd, " ");
		char *argstart = cmd + cmdlen + 1;
		/* Normal commands, may be plain command or address suffix in hex. */
		if (PREFIX(cmd, "q") || PREFIX(cmd, "quit")) {
			gtk_main_quit();
		} else if (PREFIX(cmd, "d") || PREFIX(cmd, "dump")) {
			return gidis_command_dump(prg, argstart);
		} else if (PREFIX(cmd, "a") || PREFIX(cmd, "alias")) {
			return gidis_command_alias(prg, argstart);
		} else if (PREFIX(cmd, "o") || PREFIX(cmd, "open")) {
			return gidis_command_open(prg, argstart);
		} else if (PREFIX(cmd, "c") || PREFIX(cmd, "close")) {
			return gidis_command_close(prg);
		} else if (PREFIX(cmd, "w") || PREFIX(cmd, "write")) {
			return gidis_command_write(prg, argstart);
		}

		/* Plugin commands. */
		struct cmd_item *item = _gidis_cmd_items;
		while (item != NULL) {
			if (PREFIX(cmd, item->cmd)) {
				PyObject *args = PyTuple_New(3);
				PyTuple_SetItem(args, 0, idis_python_export_program(prg));
				PyTuple_SetItem(args, 1, idis_python_export_instr(gidis_view_get_selected()));
				PyTuple_SetItem(args, 2, idis_python_export_str(argstart));
				PyObject_CallObject( item->callback.python, args);
				PyErr_Print();
				return NULL;
			}
			item = item->next;
		}
		
		/* Try address if everything else failed. */
		return gidis_command_goaddr(prg, cmd);
	} else if (command[0] == '/') {
		/* Searches */
		return gidis_command_search(prg, cmd);
	} else if (command[0] == '!') {
		/* Python commands */
		return M_ERR "Python command not supported yet.";
	} else {
		return M_ERR "Unknown prefix character.";
	}

	return NULL;
}

void gidis_add_command_hook( char *cmd, PyObject *callback ) {
	struct cmd_item *item = _gidis_cmd_items;
	_gidis_cmd_items = malloc(sizeof(struct cmd_item));
	_gidis_cmd_items->next = item;
	_gidis_cmd_items->cmd = strdup(cmd);
	_gidis_cmd_items->callback.python = callback;
}
