/******************************************************************************
 * Filename: view.c
 *
 * Created: 2012/03/14 21:42
 * Author: Ladislav Láska
 * e-mail: laska@kam.mff.cuni.cz
 *
 ******************************************************************************/

#include "../core.h"
#include "../config.h"

#include <assert.h>
#include <gtk/gtk.h>

#include "view.h"
#include "editor.h"
#include "command.h"
#include "region.h"
#include "util.h"
#include "alias.h"
#include "context_menu.h"
#include "keys.h"

#define LINEBUF_SIZE 4096

/* Column where data begin: address, opcode, mnemonics, annotations */
int _gidis_view_layout[] = { 5, 8, 23, 65, 120 };
int _gidis_view_layout_active = 0;
#define LAYOUT_JUMPARROW_START 1
//int _gidis_view_layout[] = { 5, 8, -1, 30, 100 };
char *_gidis_view_layout_names[] = { "lstart", "addr", "opcd", "mnem", "comment", NULL };
int _gidis_view_line_height = 14;

struct program *_gidis_prg;

static cairo_surface_t *viewer = NULL;
GtkWidget *_gidis_view_widget = NULL;

struct view_state _gidis_view_state;

struct instr_info *_gidis_view_instr_info;
int _gidis_view_instr_info_len;

gchar *_gidis_view_message = NULL;
gboolean _gidis_view_message_error = FALSE;

GArray *_gidis_view_search_marks = NULL;
int _gidis_view_search_pos = 0;
int _gidis_view_cursor = 0;

cairo_font_extents_t fx;
cairo_text_extents_t ex;

struct rgb {
	double r; double g; double b;
};

struct rgb styles[] = { 
	{  1,.8,  1 },
	{ .8, 1,  1 },
	{  1, 1, .8 }
};

/* Selected instruction */
struct instr *gidis_view_get_selected() {
	assert(_gidis_view_cursor >= 0);
	assert(_gidis_view_cursor < _gidis_view_state.lines);
	return _gidis_view_state.linemap[_gidis_view_cursor]->instr;
}


/* Search stuff */
int _gidis_view_search_sortfn( struct search_mark *a, struct search_mark *b ) {
	return a->info - b->info;
}

void gidis_view_search_new() {
	if (_gidis_view_search_marks->len > 0) {
		g_array_remove_range(_gidis_view_search_marks, 0, _gidis_view_search_marks->len);
	}
	_gidis_view_search_pos = 0;
}

void gidis_view_search_mark(int info, int type) {
	struct search_mark m = { .info = info, .type = type };
	g_array_append_val(_gidis_view_search_marks, m);
}

void gidis_view_search_mark_hi( int info, int type, int hi_off, int hi_start, int hi_end ) {
	struct search_mark m = { .info = info, .type = type, .hi_off = hi_off, .hi_start = hi_start, .hi_end = hi_end };
	g_array_append_val(_gidis_view_search_marks, m);
}

void gidis_view_search_go() {
	struct search_mark m = g_array_index(_gidis_view_search_marks, struct search_mark, _gidis_view_search_pos );
	/* _gidis_view_cursor = _gidis_view_instr_info[m.info].baseline; */
	gidis_view_scroll_to( m.info );
}

void gidis_view_search_first() {
	g_array_sort(_gidis_view_search_marks, (GCompareFunc) _gidis_view_search_sortfn);
	_gidis_view_search_pos = 0;
	gidis_view_search_go();
}

void gidis_view_search_next() {
	if (_gidis_view_search_marks->len <= _gidis_view_search_pos) 
		gidis_view_show_error("No search in progress.");
	else if (_gidis_view_search_pos < _gidis_view_search_marks->len-1) {
		_gidis_view_search_pos++;
		gidis_view_search_go();
	} else
		gidis_view_show_error("Already on the last entry.");
}

void gidis_view_search_prev() {
	if (_gidis_view_search_marks->len <= _gidis_view_search_pos) 
		gidis_view_show_error("No search in progress.");
	else if (_gidis_view_search_pos > 0) {
		_gidis_view_search_pos--;
		gidis_view_search_go();
	} else
		gidis_view_show_error("Already on the first entry.");
}

/* Build linebuf in format:
 * \t - switch to next column
 * \n - switch to next row in currect column
 * ^[t=x] - switch to column named x, this is: addr,opcd,mnem,comment or numerical constant
 * ^[f=fn] - where fn can be one of bold,normal,italic
 * ^[baseline] - makes current line the baseline for the rest
 * ^[a=n]...^/ - creates active area identified by n.
 */
void gidis_view_build_linebuf(struct instr_info *info) {
	struct instr *instr = info->instr;
	char linebuf[LINEBUF_SIZE] = {0};
	char *buftop = linebuf;
#define LINEBUF_FREE (LINEBUF_SIZE-(buftop-linebuf))
	if (instr != NULL) {
		/* First, look for a symbol */
		struct symbol *sym = idis_lookup_symbol( _gidis_prg, instr->vmaddr );
		if (sym != NULL) {
			buftop += snprintf(buftop, LINEBUF_FREE, "\n%016x <%s>:\n^[baseline]", (unsigned int) sym->addr, sym->name);
		}

		/* Print instruction address */
		buftop += snprintf(buftop, LINEBUF_FREE, "^[t=addr]%x:", (unsigned int) instr->vmaddr);

		/* Print raw opcode */
		buftop += snprintf(buftop, LINEBUF_FREE, "^[t=opcd]");
		for (int b = 0; b < instr->raw_length; b++) {
			if (b > 0 && (b % _config_instr_bytes_per_line) == 0)
				buftop += snprintf(buftop, LINEBUF_FREE, "\n");
			buftop += snprintf(buftop, LINEBUF_FREE, "%02x ", (unsigned int) instr->raw[b] & 0xff);
		}

		/* Print instruction mnemonics */
		char *buf = idis_instr_tostr(instr, IFMT_DEFAULT);
		char *split = " ";
		assert(buf);
		buftop += snprintf(buftop, LINEBUF_FREE, "^[t=mnem]^[f=bold]^[a=%i,%i]%-5s^/", 0, GRT_MNEMONIC, buf);
		while ((buf = idis_instr_tostr_arg()) != NULL) {
			if (_config_gidis_show_aliases) {
				struct expr *expr = idis_instr_tostr_arg_expr();
				char *alias = gidis_alias_lookup(expr);
				if (alias) buf = alias;
			}
			buftop += snprintf(buftop, LINEBUF_FREE, "%s^[a=%i,%i]%s^/", split, idis_instr_tostr_arg_index()+10, GRT_OPERAND, buf);
			split = ", ";
		}
		buftop += snprintf(buftop, LINEBUF_FREE, "^[f=normal]");

		/* Print jump symbols */
		if (instr->def->flags & IDF_BRANCH) {
			struct op *op;
			for (int j = 0; (op = instr->ops[j]) != NULL; j++) {
				if (op->type == OT_JUMP_REL) {
					struct assumption *as = idis_assumption_getfor( _gidis_prg, instr, AK_REG, R_RIP);
					vmaddr_t jaddr = as->value.imm + op->val;
					struct symbol *symb = idis_lookup_symbol_nearest( _gidis_prg, jaddr );
					if (symb != NULL) {
						vmaddr_t joff = jaddr - symb->addr;
						if (joff)	buftop += snprintf(buftop, LINEBUF_FREE, "    ^[f=italic]<%s%s>^[f=normal]", symb->name, strrelf(joff, IFMT_ATT));
						else		buftop += snprintf(buftop, LINEBUF_FREE, "    ^[f=italic]<%s>^[f=normal]", symb->name);
					}
				}
			}
		}

		/* Print comments and other userdata*/
		int comments_printed = 0;
		for (int j = 0; j < instr->userdatan; j++) {
			struct userdata *d = instr->userdata[j];
			if (d->type == UDT_COMMENT) {
				if (comments_printed > 0)
					buftop += snprintf(buftop, LINEBUF_FREE,"\n# %s", (char *) d->data);
				else
					buftop += snprintf(buftop, LINEBUF_FREE, "^[t=comment]# %s", (char *) d->data);
				comments_printed++;
			} else if (d->type == UDT_ANNOTATE_HOOK) {
				PyObject *args = PyTuple_New(3);
				PyTuple_SetItem(args, 0, idis_python_export_program(_gidis_prg));
				PyTuple_SetItem(args, 1, idis_python_export_instr(instr));
				PyTuple_SetItem(args, 2, (PyObject *) d->data);
				PyObject *ret = PyObject_CallObject((PyObject *) d->callback, args);
				if ((ret == NULL) || (! PyString_Check(ret))) {
					PyErr_Print();
					continue;
				}
				char *retc = PyString_AsString(ret);
				if (comments_printed > 0)
					buftop += snprintf(buftop, LINEBUF_FREE,"\n# %s", retc);
				else 
					buftop += snprintf(buftop, LINEBUF_FREE, "^[t=comment]# %s", retc);
				comments_printed++;
			}
		}

	} else {
		assert(info->sec);
		buftop += snprintf(buftop, LINEBUF_FREE, "\nDisassembly of section %s:\n", info->sec->name);
	}
	if (info->linebuf) free(info->linebuf);
	info->linebuf = strdup(linebuf);
#undef LINEBUF_FREE

}

int _gidis_view_arrows_compare_fn(int *a, int *b) {
	if ((*a) >= 0)
		return (*a)-(*b);
	else 
		return -(*a)-(*b);
}

struct instr_info *_gidis_view_linemap_for(struct instr *instr) {
	for (int i = 0; i < _gidis_view_state.lines; i++) {
		if (_gidis_view_state.linemap[i] && (instr == _gidis_view_state.linemap[i]->instr))
			return _gidis_view_state.linemap[i];
	}
	return NULL;
}

void gidis_view_arrows_draw(cairo_t *cr) {
#define CLINE(info) (14+(info->firstline + info->baseline)*14 - (fx.height/2) + fx.descent)
#define RGB_GREEN 0, 0.4, 0.05
	/* Setup some invariants */
	int x = _gidis_view_layout[LAYOUT_JUMPARROW_START]*(ex.width+3);
	int vspace_base = ex.width;
	int hspan_base = 10;
	
	int moused = _gidis_view_state.moused_line;
	struct instr_info *moused_info = NULL;
	if (_gidis_view_state.lines > moused) {
		/* Find selected index */
		if (_gidis_view_state.linemap[moused] != NULL) {
			while ((moused > 0) && (_gidis_view_state.linemap[moused] == _gidis_view_state.linemap[moused-1])) moused--;
			/*moused += _gidis_view_state.linemap[moused]->baseline; */
		}
		
		/* Find jumpedto, if any. jumpedfrom will be solved on-line, since we will be shrinking anyway. */
		moused_info = _gidis_view_state.linemap[moused];
	} else moused = -1;


	/* Arrow queue. The higher part of the value is arrow length, the lower part is first index */
	for (int i = 0; i < _gidis_view_state.lines; i++) {

		/* Do this only for first entry in linemap */
		if ((i > 0) && _gidis_view_state.linemap[i] == _gidis_view_state.linemap[i-1]) continue;
		if (_gidis_view_state.linemap[i] == NULL) continue;
		if (_gidis_view_state.linemap[i]->instr == NULL) continue;
		struct instr_info *info = _gidis_view_state.linemap[i];
		struct instr *instr = info->instr;

		int y = CLINE(info);

		/* Paint arrows out */
		if (instr && instr->jumpsto && instr->jumpsto->instr) {
			assert(instr->jumpsto->next == NULL);
			/* Check if this insruction will be shown as well */
			
			int adir =  (instr->jumpsto->instr->vmaddr < instr->vmaddr) ? -1 : 1;
			int s = 3*adir;
			int hspan = hspan_base*adir;
			int vspace = vspace_base;

			/* If we're moused, or pointing to moused, make more space */
			if ((moused == i) || (instr->jumpsto && moused_info && (moused_info->instr == instr->jumpsto->instr)))
				vspace *= 3;
			
			cairo_save(cr);
			if (instr->def->flags & IDF_CONDITIONAL)
				cairo_set_source_rgb(cr, 0, 0, 1);
			else
				cairo_set_source_rgb(cr, 1, 0, 0);

			cairo_move_to(cr, x, y);
			cairo_set_line_width(cr, 1.0);
			cairo_line_to(cr, vspace, y);
			cairo_line_to(cr, vspace,   y+hspan);
			cairo_stroke(cr);
			cairo_line_to(cr, vspace-s, y+hspan-s*2);
			cairo_line_to(cr, vspace+s, y+hspan-s*2);
			cairo_line_to(cr, vspace,   y+hspan);
			cairo_fill(cr);
			cairo_restore(cr);
		}

		/* Paint arrows in */
		if (instr && instr->jumpedfrom) {
			cairo_save(cr);
			cairo_set_line_width(cr, 1.0);
			cairo_set_source_rgb(cr, RGB_GREEN);
			/* Traverse list and determine jump points */
			int up = 0, down = 0;
			int vspace = vspace_base;
			int hspan = hspan_base;

			if (((moused == i) && (instr->jumpedfrom != NULL)) 
				|| (moused_info && moused_info->instr && moused_info->instr->jumpsto && moused_info->instr->jumpsto->instr == instr))
				vspace *= 3;

			struct instr_llist *li = instr->jumpedfrom;
			while (li != NULL) {
				if (li->instr) {
					if (li->instr->vmaddr < instr->vmaddr) up++;
					else down++;
				}
				li = li->next;
			}

			if (up || down) {
				int s = _config_gidis_arrow_size;
				cairo_move_to(cr, x, y);
				cairo_line_to(cr, x-2*s, y-s);
				cairo_line_to(cr, x-2*s, y+s);
				cairo_line_to(cr, x, y);
				cairo_fill(cr);
				cairo_move_to(cr, x, y);
				cairo_line_to(cr, vspace, y);
				if (up)
					cairo_line_to(cr, vspace, y-hspan);
				if (down)
					cairo_line_to(cr, vspace, y+hspan);
				cairo_stroke(cr);
			}
			cairo_restore(cr);
		}
	}

	/* Draw full, joined lines. */
	if (!moused_info) return;
	struct instr *instr = moused_info->instr;
	if (!instr) return;
	int vspace = vspace_base*3;
	int h = gtk_widget_get_allocated_height(_gidis_view_widget);

	cairo_save(cr);
	/* Arrows out */
	if (instr->jumpsto && instr->jumpsto->instr) {
		struct instr *t = instr->jumpsto->instr;
		struct instr_info *t_info = _gidis_view_linemap_for(t);
		if (instr->def->flags & IDF_CONDITIONAL)
			cairo_set_source_rgb(cr, 0, 0, 1);
		else
			cairo_set_source_rgb(cr, 1, 0, 0);
		int hspan = (instr->vmaddr < t->vmaddr) ? hspan_base : - hspan_base;
		cairo_move_to(cr, vspace, CLINE(moused_info)+hspan);
		if (t_info!= NULL) {
			cairo_set_line_width(cr, 1.0);
			cairo_line_to(cr, vspace, CLINE(t_info)-hspan);
		} else {
			cairo_set_line_width(cr, 0.5);
			cairo_line_to(cr, vspace, (hspan < 0) ? 0 : h);
		}
		cairo_stroke(cr);
	}

	/* Arrows in */
	struct instr_llist *li = instr->jumpedfrom;
	while (li != NULL) {
		struct instr *t = li->instr;
		struct instr_info *t_info = _gidis_view_linemap_for(t);
		cairo_set_source_rgb(cr, RGB_GREEN);
		int hspan = (instr->vmaddr < t->vmaddr) ? hspan_base : - hspan_base;
		cairo_move_to(cr, vspace, CLINE(moused_info)+hspan);
		if (t_info != NULL) {
			cairo_set_line_width(cr, 1.0);
			cairo_line_to(cr, vspace, CLINE(t_info)-hspan);
		} else {
			cairo_set_line_width(cr, 0.5);
			cairo_line_to(cr, vspace, (hspan < 0) ? 0 : h);
		}
		cairo_stroke(cr);
		li = li->next;
	}
	cairo_restore(cr);

	return;

#undef CLINE
}

static gboolean _gidis_redraw_view(GtkWidget *widget, cairo_t *cr, gpointer data) {
#define font_bold() \
	cairo_select_font_face(cr, "DejaVu Sans Mono", CAIRO_FONT_SLANT_NORMAL, CAIRO_FONT_WEIGHT_BOLD)
#define font_normal() \
	cairo_select_font_face(cr, "DejaVu Sans Mono", CAIRO_FONT_SLANT_NORMAL, CAIRO_FONT_WEIGHT_NORMAL)
#define font_italic() \
	cairo_select_font_face(cr, "DejaVu Sans Mono", CAIRO_FONT_SLANT_ITALIC, CAIRO_FONT_WEIGHT_NORMAL)
#define LINE(n) (14+14*(n-loff))


	/* Get font metrics */
	cairo_text_extents(cr, "_", &ex);
	int cl = ex.width+3;
	
	cairo_font_extents(cr, &fx);

	/* Count message box size */
	int msg_lines = 1;
	if ((!gidis_editor_is_active()) && (_gidis_view_message != NULL)) {
		/* Count lines */
		for (int i = 0; _gidis_view_message[i] != 0; i++) 
			if (_gidis_view_message[i] == '\n') msg_lines++;
	}


	/* Get canvas metrics */
	int h = gtk_widget_get_allocated_height(widget);
	int w = gtk_widget_get_allocated_width(widget);
	int lines = h/14-msg_lines;

	/* Prepare canvas */
	cairo_set_source_surface(cr, viewer, 0, 0);
	cairo_paint(cr);
	cairo_set_source_rgb(cr, 0.1, 0.1, 0.1);
	cairo_set_font_size(cr, 13);
	font_normal();

	/* Draw cursor line */
	if (_gidis_prg) {
		if (_gidis_view_cursor >= lines) 
			_gidis_view_cursor = lines-2;
		cairo_save(cr);
		cairo_set_source_rgb(cr, .5, .5, .5);
		cairo_rectangle(cr, 0, 14+14*_gidis_view_cursor-fx.height, gtk_widget_get_allocated_width(widget), fx.height+fx.descent+1);
		cairo_fill(cr);
		cairo_restore(cr);
	}

	/* Prepare linemap */
	if (lines != _gidis_view_state.lines) {
		struct instr_info **newmap = realloc( _gidis_view_state.linemap, sizeof(struct instr_info*)*lines);
		assert(newmap);
		_gidis_view_state.lines = lines;
		_gidis_view_state.linemap = newmap;
	}
	memset(_gidis_view_state.linemap, 0, sizeof(struct instr_info*)*lines);

	/* Get current search index. */
	int search_index = 0;

	/* Process data entry-by-entry */
	struct view_state tstate = _gidis_view_state;
	int loff = tstate.step;		/* Line offset to enable scrolling by-line */
	for (int i = 0; i < lines+loff; i++) {
		int li = i;		/* Line. This is current line that is drawn. */
		int di = i; 	/* Baseline. This is line where the cursor is set to on \t */
		int ld = 0;		/* LineDepth. This is number of lines needed. */
		int index = tstate.top;
		int column = 0;

		if (index >= _gidis_view_instr_info_len) break;
		struct instr_info *info = &_gidis_view_instr_info[index];
		assert(info);
		char *linebuf = info->linebuf;

		struct region *region;

		/* Draw background */
		if (info->instr && info->instr->style) {
			int index = info->instr->style % (sizeof(styles)/sizeof(struct rgb));
			double r = styles[index].r, g = styles[index].g, b = styles[index].b;
			if (li+info->baseline-loff == _gidis_view_cursor) {
				r*=2; g*=2; b*=2;
			}
			gidis_util_fill_color(cr, 0, LINE(li+info->baseline)-fx.height, w, LINE(li+info->baseline)+fx.descent+1, r, g, b);
		}

		/* Draw stuff from linebuf */
		cairo_move_to(cr, 1, LINE(li));
		int ci = 0;
		while (linebuf[ci] != 0) {
			/* Handle meta characters */
			if (linebuf[ci] == '\t') {
				cairo_move_to(cr, _gidis_view_layout[++column]*cl, LINE(li));
				li = di;
				ci++;
				continue;
			} else if (linebuf[ci] == '\n') {
				newline:
				cairo_move_to(cr, _gidis_view_layout[column]*cl, LINE(++li));
				if ((_gidis_view_layout[column] >= 0) && (li - i > ld)) ld = li - i;
				ci++;
				continue;
			} else if (strncmp(linebuf+ci, "^/", 2) == 0) {
				assert(region);
				gidis_region_end(cr, region);
				gidis_region_draw(cr, info, region);
				region = NULL;
				ci += 2;
			} else if ((linebuf[ci] == '^') && (linebuf[ci+1] == '[')) {
				if ((strncmp(linebuf+ci, "^[t=", 4) == 0) 
				|| (strncmp(linebuf+ci, "^[f=", 4) == 0) 
				|| (strncmp(linebuf+ci, "^[a=", 4) == 0)) {
					size_t arglen = strcspn(linebuf+ci+4, "]");
					char *arg = linebuf+ci+4;
					switch ((linebuf+ci)[2]) {
						case 't':;
							for (int nc = 0; _gidis_view_layout_names[nc] != NULL; nc++) {
								if (strncmp(_gidis_view_layout_names[nc], arg, arglen) == 0) {
									li = di;
									column = nc;
									cairo_move_to(cr, _gidis_view_layout[column]*cl, LINE(li));
									break;
								}
							}
							break;
						case 'f':
							if (strncmp(arg, "bold", arglen) == 0) {
								font_bold();
							} else if (strncmp(arg, "italic", arglen) == 0) {
								font_italic();
							} else if (strncmp(arg, "normal", arglen) == 0) {
								font_normal();
							} else 
								warn("Invalid font style.");
							break;
						case 'a':;
							char *cnv_end;
							long int argint = strtol(arg, &cnv_end, 10);
							long int arg2int = 0;
							if (cnv_end[0] == ',') {
								char *cnv_end2;
								arg2int = strtol(cnv_end+1, &cnv_end2, 10);
								cnv_end = cnv_end2;
							}
							if (cnv_end[0] != ']') {
								warn("Invalid active identifier.");
								break;
							}
							region = gidis_region_byid(info, argint);
							if (region == NULL)
								region = gidis_region_new(info, argint);
							region->type = arg2int;
							assert(region);
							gidis_region_start(cr, region);
							break;
						default:
							warnf("Invalid directive %c.", (linebuf+ci)[2]);
					}
					ci += arglen+5;
					continue;
				} else if (strncmp(linebuf+ci, "^[baseline]", 11) == 0) {
					di = li;
					ci += 11;
				} else {
					warn("Invalid escape sequence.");
					ci++;
				}
			}
			/* Print other stuff. First, get the position and space we have left. 
			 * Then print what fits and do linebreak. */
			double x,y;
			cairo_get_current_point(cr, &x, &y);
			int space = (w-x)/(cl) - 2;
			int dobreak = 0;
			size_t tlen = strcspn(linebuf+ci, "\t\n^");
			if (tlen > space) { tlen = space; dobreak = 1; }
			char *tstart = linebuf+ci;
			char nc = tstart[tlen];
			tstart[tlen] = 0;
			if ((li-loff >= 0) && (li-loff < lines) && (_gidis_view_layout[column] >= 0)) {
				cairo_show_text(cr, tstart);
			} else dobreak = 0;
			tstart[tlen] = nc;
			ci += tlen;
			if (dobreak) {
				ci--;
				goto newline;
			}
		}
		
		/* Draw search highlights */
		struct search_mark m;
		while (_gidis_view_search_marks 
			&& _gidis_view_search_marks->len > search_index
			&& (m = g_array_index(_gidis_view_search_marks, struct search_mark, search_index)).info <= index
		) {
			if (m.info < index) { search_index++; continue; }
			search_index++;

			if (m.type == GST_FULLTEXT) {
				int l = LINE(i+m.hi_off);
				gidis_util_invert_color(cr, (cl)*m.hi_start, l-fx.ascent-2, (cl)*(m.hi_end), l+fx.descent+1, 0, .5, .5);
			}
			else break;
		}

		/* Advance state counter. */
		tstate.top++;
		tstate.step = 0;
		/* Update lines, which were guessed at the beginning and may not be true now */
		info->lines = ld+1;
		info->baseline = di - i;
		info->firstline = i-loff;
		for (int l = 0-loff; l < info->lines-loff; l++) {
			if (l+i < 0) continue;
			/* Check if we are past; some lines could have been read, but not painted. */
			if (l+i>= _gidis_view_state.lines) break;
			_gidis_view_state.linemap[l+i] = info;
		}
		i += ld;
	}

	gidis_view_arrows_draw(cr);

	loff = 0; /* Used by LINE macro */
	if (gidis_editor_is_active()) {
		gidis_util_add_color(cr, 0, LINE(lines)-fx.height, w, h, .8, .8, .8);
		_gidis_editor_draw(cr, 1, LINE(lines));
	} else
	if (_gidis_view_message != NULL) {
		cairo_save(cr);
		cairo_save(cr);
		gidis_util_add_color(cr, 0, LINE(lines)-fx.height, w, h, .8, .8, .8);
		cairo_restore(cr);
		if (_gidis_view_message_error) {
			cairo_set_source_rgb(cr, 1.0, 0.0, 0.0);
			font_bold();
		}
		char *buf = _gidis_view_message;
		int l = 0;
		while (buf[0] != 0) {
			size_t end = strcspn(buf, "\n");
			char end_char = buf[end];
			buf[end] = 0;
			cairo_move_to(cr, 1, LINE(lines+l));
			cairo_show_text(cr, buf);
			buf[end] = end_char;
			buf += end+1;
			l++;
			assert(l<=msg_lines);
			if (end_char == 0) break;
		}
		cairo_restore(cr);
	}
	
	return TRUE;
}

static void _gidis_clear_view() {
	cairo_t *cr = cairo_create(viewer);
	cairo_set_source_rgb(cr, 1, 1, 0.5);
	cairo_paint(cr);
	cairo_destroy(cr);
}

void gidis_view_show_message_enc(gchar *message) {
	if (message[0] == 0x1b)
		gidis_view_show_error(message+1);
	else
		gidis_view_show_message(message);
}

void _gidis_view_show_message(gchar *message, gboolean error) {
	if (_gidis_view_message != NULL) g_free(_gidis_view_message);
	_gidis_view_message = g_strdup(message);
	_gidis_view_message_error = error;
	gtk_widget_queue_draw(_gidis_view_widget);
}

void gidis_view_hide_message() {
	if (_gidis_view_message != NULL) {
		g_free(_gidis_view_message);
		_gidis_view_message = NULL;
	}
}

void gidis_view_show_message(gchar *message) {
	_gidis_view_show_message(message, FALSE);
}

void gidis_view_show_error(gchar *message) {
	_gidis_view_show_message(message, TRUE);
}

static gboolean _gidis_resize_view(GtkWidget *widget, GdkEventConfigure *event, gpointer data) {
	if (viewer)
		cairo_surface_destroy(viewer);
	
	/* Initialize canvas */
	viewer = gdk_window_create_similar_surface(gtk_widget_get_window(widget),
			CAIRO_CONTENT_COLOR,
			gtk_widget_get_allocated_width(widget),
			gtk_widget_get_allocated_height(widget));

	_gidis_clear_view();
	
	return TRUE;
}

void gidis_view_scroll_to( int index ) {
	struct instr_info *info = &_gidis_view_instr_info[index];
	_gidis_view_state.top = index;
	_gidis_view_state.step = 0;
	gtk_widget_queue_draw(_gidis_view_widget);
	int done = _gidis_do_scroll(_gidis_view_widget, _gidis_view_cursor - info->baseline);
	_gidis_view_cursor = done+ info->baseline;
}

void gidis_view_scroll_to_instr( struct instr *instr ) {
	for (int i = 0; i < _gidis_view_instr_info_len; i++) {
		if (_gidis_view_instr_info[i].instr == instr) {
			/* This is it. Scroll there. */
			gidis_view_scroll_to( i );
			break;
		}
	}

}

int _gidis_do_scroll(GtkWidget *widget, int amount) {
	int done = 0;
	if (_gidis_prg == NULL) return 0;
	while (amount != 0) {
		done++;
		if (amount > 0) {
			amount--;
			_gidis_view_state.step--;
			if (_gidis_view_state.step < 0) {
				if (_gidis_view_state.top == 0) { 
					_gidis_view_state.step = 0; done--; 
				} else {
					_gidis_view_state.top--;
					_gidis_view_state.step = _gidis_view_instr_info[_gidis_view_state.top].lines-1;
				}
			}
		} else {
			amount++;
			_gidis_view_state.step++;
			if (_gidis_view_state.step >= _gidis_view_instr_info[_gidis_view_state.top].lines) {
				if (_gidis_view_state.top == _gidis_view_instr_info_len-1) {
					_gidis_view_state.step--;
					done--;
				} else {
					_gidis_view_state.top++;
					_gidis_view_state.step = 0;
				}
			}
		}
	}
	
	gtk_widget_queue_draw(widget);
	return done;
}

static gboolean _gidis_scroll_view(GtkWidget *widget, GdkEventScroll *event, gpointer data) {
	gidis_view_hide_message();
	switch (event->direction) {
		case GDK_SCROLL_UP:
			_gidis_do_scroll(widget, 2);
			break;
		case GDK_SCROLL_DOWN:
			_gidis_do_scroll(widget, -2);
			break;
		default:;
	}

	return TRUE;
}

static gboolean _gidis_click_view(GtkWidget *widget,GdkEventButton *event, gpointer data) {
	int line = event->y / _gidis_view_line_height;
	struct instr_info *info = _gidis_view_state.linemap[line];
	switch (event->type) {
		case GDK_BUTTON_PRESS:;
			if (event->button == 3) {
				struct region *r = gidis_region_at(event->x, event->y);
				if (r != NULL)
					gidis_context_menu_popup(info, r);
			}
			if (event->button == 1) {
				_gidis_view_cursor = line;
			}
			break;
		case GDK_2BUTTON_PRESS:;
			if (event->button == 1) {
				_gidis_view_cursor = line;
				if (info && info->instr && info->instr->jumpsto && info->instr->jumpsto->instr)
					gidis_view_scroll_to_instr(info->instr->jumpsto->instr);
			}
			break;
		default:
			return FALSE;
	}
	
	gtk_widget_queue_draw(widget);

	return TRUE;
}

void gidis_view_clear_style() {
	for (int i = 0; i < _gidis_view_instr_info_len; i++) {
		struct instr_info *info = &_gidis_view_instr_info[i];
		if (info->instr)
			info->instr->style = 0;
	}
}

void gidis_view_toggle_style() {
	if (_gidis_view_cursor < 0) return;
	struct instr *instr = _gidis_view_state.linemap[_gidis_view_cursor]->instr;
	if (instr == NULL) return;
	instr->style = (instr->style + 1) % 2;
}

void gidis_view_toggle_bytes() {
	_gidis_view_layout_active = (_gidis_view_layout_active + 1) % 2;

	for (int i = 0; i < (sizeof(_gidis_view_layout)/sizeof(int)); i++) {
		_gidis_view_layout[i] = _config_gidis_view_layouts[_gidis_view_layout_active][i];

	}
}

static gboolean _gidis_motion_handler(GtkWidget *widget, GdkEventMotion *event, gpointer data) {
	int line = event->y / _gidis_view_line_height;
	if (line >= _gidis_view_state.lines) line = _gidis_view_state.lines-1;
	if (line != _gidis_view_state.moused_line) {
		_gidis_view_state.moused_line = line;
		gtk_widget_queue_draw(widget);
	}

	return TRUE;
}

void gidis_view_rebuild_linebuf() {
	debug("Rebuilding linebuf.");
	for (int i = 0; i < _gidis_view_instr_info_len; i++)
		gidis_view_build_linebuf(&_gidis_view_instr_info[i]);
}

void gidis_view_prg_load(struct program *prg) {
	_gidis_prg = prg;
	
	gidis_alias_init();

	/* Read instructions and create index */
	for (int phase = 0; phase < 2; phase++) {
		int k = 0;
		for (int i = 0; prg->sections[i] != NULL; i++) {
			if ((prg->sections[i]->flags & SF_EXE) == 0) continue;
			/* Create header for section */
			if (phase == 1) {
				_gidis_view_instr_info[k].instr = NULL;
				_gidis_view_instr_info[k].sec = prg->sections[i];
				_gidis_view_instr_info[k].lines = 3;
				gidis_view_build_linebuf(&_gidis_view_instr_info[k]);
			}
			k++;
			/* Create entry for every instruction */
			for (int j = 0; j < prg->sections[i]->instrn; j++) {
				if (phase == 1) {
					_gidis_view_instr_info[k].instr = prg->sections[i]->instrs[j];
					_gidis_view_instr_info[k].sec = prg->sections[i];
					_gidis_view_instr_info[k].lines = 1;
					gidis_view_build_linebuf(&_gidis_view_instr_info[k]);
				}
				k++;
			}
		}
		if (phase == 0) {
			_gidis_view_instr_info_len = k;
			_gidis_view_instr_info = malloc(sizeof(struct instr_info)*_gidis_view_instr_info_len);
			memset(_gidis_view_instr_info, 0, sizeof(struct instr_info)*_gidis_view_instr_info_len);
			k = 0;
		}
	}


	/* Initialize _gidis_view_state */
	memset(&_gidis_view_state, 0, sizeof(struct instr_info));

	_gidis_view_search_marks = g_array_new(TRUE, TRUE, sizeof(struct search_mark));
}

void gidis_view_prg_free() {
	gidis_alias_free();
	g_array_free( _gidis_view_search_marks, TRUE );
	_gidis_view_search_marks = NULL;

	/* Free instrinfo structures */
	for (int i = 0; i < _gidis_view_instr_info_len; i++) {
		if (_gidis_view_instr_info[i].linebuf)
			free(_gidis_view_instr_info[i].linebuf);
		struct region *r = _gidis_view_instr_info[i].regionlist;
		struct region *next;
		while (r) {
			next = r->next;
			gidis_region_free(r);
			r = next;
		}

	}

	free(_gidis_view_instr_info);
	_gidis_view_instr_info = NULL;
	_gidis_prg = NULL;
	_gidis_view_instr_info_len = 0;
}

GtkWidget *gidis_view_new(struct program *prg) {
	assert(viewer == NULL);

	if (prg)
		gidis_view_prg_load(prg);

	/* Setup drawing area */
	GtkWidget *da;
	da = gtk_drawing_area_new();
	gtk_widget_set_can_focus(da, TRUE);
	g_signal_connect(da, "draw", G_CALLBACK(_gidis_redraw_view), NULL);
	g_signal_connect(da, "configure-event", G_CALLBACK(_gidis_resize_view), NULL);
	g_signal_connect(da, "scroll-event", G_CALLBACK(_gidis_scroll_view), NULL);
	g_signal_connect(da, "button-press-event", G_CALLBACK(_gidis_click_view), NULL);
	g_signal_connect(da, "key-press-event", G_CALLBACK(gidis_key_handler), NULL);
	g_signal_connect(da, "motion-notify-event", G_CALLBACK(_gidis_motion_handler), NULL);
	gtk_widget_add_events(da, GDK_SCROLL_MASK);
	gtk_widget_add_events(da, GDK_KEY_PRESS_MASK);
	gtk_widget_add_events(da, GDK_KEY_RELEASE_MASK);
	gtk_widget_add_events(da, GDK_BUTTON_PRESS_MASK);
	gtk_widget_add_events(da, GDK_BUTTON_RELEASE_MASK);
	gtk_widget_add_events(da, GDK_POINTER_MOTION_MASK);
	gtk_widget_set_size_request(da, 100, 100);
	_gidis_view_widget = da;
	gidis_view_toggle_bytes(); /* Apply configuration. */
	gidis_view_toggle_bytes();
	return da;
}

