#!/usr/bin/perl

use XML::LibXML;
use Data::Dumper;
use Switch;

my $sf = ($ARGV[0]) ? $ARGV[0] : "x86reference.xml";
my $ninstr = 0;

my $instr;

# Instruction array 
my %ia = ("0" => {}, "0F" => {});

my $parser = XML::LibXML->new();
my $doc = $parser->parse_file( $sf ) or die("Could not parse file "+$sf+"\n");

my $xc = XML::LibXML::XPathContext->new($doc);

open H, "+>instr_data.c" or die("Could not open output file!");

# Process one-byte instructions (no prefix)
for my $node (($xc->findnodes("/x86reference/one-byte/pri_opcd"))) {
	do_pri_opcd("", $node);
}

# Process multiple (prefix 0x0F)
for my $node (($xc->findnodes("/x86reference/two-byte/pri_opcd"))) {
	do_pri_opcd("0F", $node);
}


printf "Found %i opcodes/instructions.\n", $ninstr;

dump_instructions();

close H;

sub do_pri_opcd {
	my ($prefix_0f, $node) = @_;
	die ("Not a primary opcode: '".$node->getName()."'!") if ($node->getName() ne "pri_opcd");
#	printf "%s\n",  $node->getName();
	my $opcode = $node->getAttribute("value");
	# Have the pri_opcd object. This has multiple entries, some of them may be
	# marked as invalid in 64bit mode (<entry attr="invd" mode="e">)
	my $xcp = XML::LibXML::XPathContext->new($node);

	# Check for instructions invalid in 64bit mode (pri_opcd has entry with attr=invd & mode=e)
	my $invd = ($xcp->findnodes("./entry[(\@attr='invd' and \@mode='e')]"));
	if ($invd) {
		next;
	}

	my $is_null = 0;
	# Handle null prefixes
	for my $entry (($xcp->findnodes("./entry[(\@attr='null' and \@mode='e')]"))) {
		do_instruction($prefix_0f, $opcode, $node, $entry, $xcp);
		$is_null++;
	}
	next if ($is_null);

	# Give precedence to mode='e' instructions...
	#$modx = ($xcp->findnodes("./entry[\@mode='e']"));
	for my $entry (($xcp->findnodes("./entry"))) {
		do_instruction($prefix_0f, $opcode, $node, $entry, $xcp);
	}
}

sub do_instruction {
	my ($prefix_0f, $opcode, $node, $entry, $xcp) = @_;
	die ("Not an instruction entry: '".$entry->getName()."'!") if ($entry->getName() ne "entry");
	
	my $x = XML::LibXML::XPathContext->new($entry);

	# In case of multibyte instruction (0x0F prefix), 0x38 and 0x3A primary opcode is prefix for secondary opcode
	if (($opcode =~ /^(38|3A)/) && ($prefix_0f eq "0F")) {
	; # Probably doesn't matter, it is handled well.
	}

	my $opcode2 = $x->findvalue( "sec_opcd" );
	my $prefix= $x->findvalue( "pref" );
	my $mnem = $x->findvalue( "syntax[1]/mnem" );
	my $desc = $x->findvalue( "note/brief" );
	my $ext = "8";
	if ($x->findnodes("opcd_ext")) {
		$ext = $x->findvalue( "opcd_ext" );
	}
	if (($x->findvalue("grp1") eq "x87fpu") 
		or ($x->findvalue("grp2") eq "x87fpu") 
		or ($x->findvalue("grp3") eq "x87fpu") 
		or ($x->findvalue("grp2") =~ "segreg") 
	)
	{
		next;
	}

# Check if this instruction should be skipped due to 64bit override (another instruction with mode=e
	if (($ext ne "8") and ($entry->getAttribute("mode") ne "e") and (($xcp->findnodes("./entry[\@mode='e']/opcd_ext[text() = \"$ext\"]")))) {
		next;
	} elsif (($ext eq "8") and ($entry->getAttribute("mode") ne "e") and (($xcp->findnodes("./entry[\@mode='e']")))) {
		next;
	}

	# Skip instruction extensions
	my $iext = $x->findvalue("instr_ext");
	if ($iext ne "") {
		next;
	}

	# Check for prefixes, these are not auto-generated.
	my $isprefix = 0;
	if ($x->findvalue("grp1") eq "prefix") {
		$isprefix = 1;
	}

	# Check if it is stil valid
	if ($x->exists("proc_end") && $x->findvalue( "proc_end" ) < 10) {
		next;
	}

	# Check for undefined instructions
	if ($entry->getAttribute("attr") eq "undef") {
		next;
	}

	# Check for NULL prefixes
	if ($entry->getAttribute("attr") eq "null") { 
		$mnem = "_NULL_PF"; 
	}

	# Check wether instruction is (un)conditional jump
	my $branch = 0;
	$branch = 1 if ($x->findvalue("grp2") =~ "branch");
	$branch = 2 if (($x->findvalue("grp3") =~ "cond") && $branch);
	my $near = 0;
	$near = 1 if ($x->findvalue("grp2") =~ "near");
	my $far = 0;
	$far = 1 if ($x->findvalue("grp2") =~ "far");

	my $dir = $entry->getAttribute("direction");

	$opcode2 ||= "0";
	$prefix ||= "0";
	$prefix_0f ||= "0";



	if ($isprefix) {
		$mnem ||= "_NO_MNEM";
		next;
	} else {
		my @variants = $x->findnodes("./syntax[not(\@redundant)]");
		#nomem vs. mem

		for $variant (@variants) {
			my $flags = 0;
			# Add branch flags
			$flags .= " | IDF_BRANCH " if ($branch);
			$flags .= " | IDF_CONDITIONAL" if ($branch == 2);
			$flags .= " | IDF_JUMP_NEAR" if ($branch and $near);
			$flags .= " | IDF_JUMP_FAR"  if ($branch and $far);

			my @ops = ();
			my $mask = "ff";
			my %ni = (
				"mnem" => $mnem, 
				"desc" => $desc,
				"prefix" => $prefix,
				"opcode" => $opcode, 
				"opcode2" => $opcode2, 
				"extension" => $ext,
				"flags" => $flags,
				"mask" => "ff",
				"ops" =>\@ops
			);
			
			my $ax = XML::LibXML::XPathContext->new($variant);
			my $nonbyte = 0;
			my $opn = -1;
			for my $arg (($ax->findnodes( "./*" ))) {
				my $argx = XML::LibXML::XPathContext->new($arg);
				if ($arg->getName() eq "mnem") {
					$ni{"mnem"} = $argx->findvalue(".");
					next;
				}
				$opn++;
				my $a = $argx->findvalue("a") || "def";
				my $t = $argx->findvalue("t") || "def";
				my $d = 0;
				my $r = 0;
				my $ft = undef;
				$ni{"mask"} = "f8" if ($a eq "Z");
				if ($a eq "def") {
					if ($arg->hasAttribute("nr")) {
						my $nr = $arg->getAttribute("nr");
						my $group = $arg->getAttribute("group");
						$t = $arg->getAttribute("type") || "FAIL";
						$t = "def" if ($group eq "x87fpu");
						if ($group eq "msr")
						{ $nr = "111"; $t = "def"; } 
						$t = "def" if ($group eq "ctrl"); #only CLTS
						$t = "pi" if ($group eq "mmx");
						$t = "dq" if ($group eq "xmm");
						$a = "REG";
						$d = $nr;
					} elsif ($arg->hasAttribute("address")) {
						$nt = $arg->getAttribute("address");
						switch ($nt) {
							case "I"
								{
									$t = "CONST";
									$d = $argx->findvalue(".");
								}
							case "SC"
								{
									$t = "STACK";
									$ni{"flags"} .= " | IDF_64BIT";
									# prefix 66 -> 16bit
								}
							else { $t = "def"; }
						}
					} else {
						my $v = $argx->findvalue(".");
						printf("Predefined valued, but no argument? $v\n");
				
						switch ($v) {
							case "Flags" {}
							case "eFlags" {}
						}
					}
				}
				$a = "A_".$a;
				$ft ||= "T_".$t;

				# Mark this instruction that it has non-byte argument
				$nonbyte++ if ($ft ne "T_b");
				
				my $opflags = 0;
				$opflags .= " | OF_DEST" if ($arg->getName() eq "dst");					# OF_DEST
				$opflags .= " | OF_HIDDEN" if ($arg->getAttribute("displayed") eq "no"); # OF_HIDDEN

				# If direction is 0, E is destination, G otherwise
				if ($entry->hasAttribute("direction")) {
					$opflags .= " | OF_VDEST" if ($a eq "A_E" and $dir == 0);
					$opflags .= " | OF_VDEST" if ($a eq "A_G" and $dir == 1);
				} elsif (($mnem eq "CMP") and (($a eq "A_E") or ($a eq "A_G") or ($a eq "A_REG"))) {
				# CMP instructions with immediate does not have direciton flag, however the virtual
				# destination is always the register/memory
					$opflags .= " | OF_VDEST";
				}
				
				push @ops, { 
					"flags" => $opflags,
					"addr" => $a,
					"type" => $ft,
					"def" => $d
				};
			}
			
			# Mark instructions that have only
			$ni{"flags"} .= " | IDF_8BIT" if ($nonbyte == 0);

				
			# If addressing is Z, it means 3 least significant bits specify GPR
			next if (($ni{"mask"} eq "f8") and ($opcode == 90));

			$ia{$prefix_0f}->{$opcode} = [] if not @ia{$prefix_0f}->{$opcode};
			push @{$ia{$prefix_0f}->{$opcode}}, \%ni;

		}
	}


	$ninstr++;
}

sub dump_instructions {
	my $in = 0;
	# Initialize hash of opcodes
	my %opcds = {};
	my %opcds_0f = {};
	foreach $n (0..255) {
		$opcds{$n} = [];
		$opcds_0f{$n} = [];
	}
	#print Dumper(\%ia);
	
	printf H "#include \"instr_data.h\"\n";

	
	# Generate and expand opcodes
	foreach $prefix_0f (keys %ia) {
		foreach $opcd (sort keys %{$ia{$prefix_0f}}) {
			foreach $instr (@{$ia{$prefix_0f}{$opcd}}) {
				my %h = %$instr;
				my @ops = ();
				# Create object for each option.
				foreach $op (@{$h{"ops"}}) {
					my %o = %$op;
					printf H "struct op_def __obj_def_$in = { .flags = %s, .addr = %s, .type = %s, .def = %s };\n", 
						$o{"flags"}, $o{"addr"}, $o{"type"}, $o{"def"};
					push @ops, "__obj_def_".$in++;
				}
				# Print list of options
				print H "struct op_def *__obj_def_$in [] = { ";
				foreach $obj (@ops) {
					print H "&$obj, "
				}
				print H "NULL };\n";
				my $opa = "__obj_def_".$in++;
				# Print instruction struct, it may be needed to expand it according to mask!
				my $stepto = ($h{"mask"} eq "f8") ? 7 : 0;
				for $mask (0 .. $stepto) {
					printf H "struct instr_def __obj_def_$in = {\n"
						."\t.mnem = \"%s\", \n"
						."\t.desc = \"%s\", \n"
						."\t.prefix = 0x%02s, \n"
						."\t.opcd = 0x%02x,\n"
						."\t.opcd2 = 0x%02s,\n"
						."\t.extension = %i,\n"
						."\t.flags = %s,\n"
						."\t.mask= 0x%02s,\n"
						."\t.ops = $opa \n};\n",
						$h{"mnem"}, $h{"desc"}, $h{"prefix"}, (hex $h{"opcode"}) | $mask, 
						$h{"opcode2"}, $h{"extension"}, $h{"flags"}, $h{"mask"};
					my $obj = "__obj_def_".$in++;
					# Push instruction to appropriate array
					my $opcd_n = hex $h{"opcode"};
					$opcd_n |= $mask;
					if ($prefix_0f eq "0F") {
						push @{$opcds_0f{$opcd_n}}, $obj;
					} else {
						push @{$opcds{$opcd_n}}, $obj;
					}
				}
			}
		}
	}

	# Create arrays for all one-byte opcodes
	foreach $n (0..255) {
		printf H "struct instr_def *__opcd_def_%02x[] = {\n", $n;
		foreach $instr (@{$opcds{$n}}) {
			printf H "\t&$instr,";
		}
		printf H "\tNULL \n};\n";
	}

	# Create arrays for all two-byte opcodes
	foreach $n (0..255) {
		printf H "struct instr_def *__opcd_def_0f%02x[] = {\n", $n;
		foreach $instr (@{$opcds_0f{$n}}) {
			printf H "\t&$instr,";
		}
		printf H "\tNULL \n};\n";
	}

	print H "struct instr_def **_opcds[] = {\n";
	foreach $n (0..255) {
		printf H "\t__opcd_def_%02x, \n", $n;
	}
	print H "NULL \n};\n";

	print H "struct instr_def **_opcds0f[] = {\n";
	foreach $n (0..255) {
		printf H "\t__opcd_def_0f%02x, \n", $n;
	}
	print H "\tNULL \n};\n";


}
